from __future__ import print_function
import warnings
import matplotlib.colors as mcolors
import six
import logging
import numpy as np
import ifigure
from ifigure.utils.cbook import ProcessKeywords
from collections import OrderedDict
from functools import wraps
from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
import ifigure.utils.debug as debug
import traceback
'''
   BookViewrInteractive

    A class which implement interactive commands for
    bookviewer
      addpage, isec, hold

    doc_string of these methods are copied from
    ifigure.interactive, and therefore not written here.
'''

'''
   most of function uses one of these decorators

      1) allow_interactive_call
         this is for plot command
         returned object from method will be added
         process update and hold keyword
         copy doc_string from ifigure.interactive

      2) allow_interactive_call2
         this is for commands to edit properties, such as title
         process update
         copy doc_string from ifigure.interactive

      3) share_doc_string
         this one just copy doc_string from ifigure.interactive

   decorator to update, hold functionality to
   methods for interactive ploting,
   also it copies doc_string from functions
   in ifigure.interactive
'''


_type_numbers = ['int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32',
                 'uint64', 'float16', 'float32', 'float64', 'complex64', 'complex128', ]
_dtype_numbers = []
for x in _type_numbers:
    try:
        _dtype_numbers.append(np.dtype(x))
    except:
        print('numpy does not know type = '+x)


class NoPageError(Exception):
    pass


class InteractiveCommandError(Exception):
    pass


dprint1, dprint2, dprint3 = debug.init_dprints('BookViewerInteractive')


def has_plot(figaxes):
    from ifigure.mto.fig_axes import FigAxes
    from ifigure.mto.fig_text import FigText
    from ifigure.mto.fig_box import FigBox
    from ifigure.mto.fig_circle import FigCircle
    from ifigure.mto.fig_legend import FigLegend
    for obj in figaxes.walk_tree():
        if (not isinstance(obj, FigAxes) and
            not isinstance(obj, FigText) and
            not isinstance(obj, FigBox) and
                not isinstance(obj, FigCircle)):
            return True
    return False


def pop_metadata(kargs):
    '''
       metadata is either dictionary
       or
           longname,
           xlongname
           ylongname and so on

      example :
           plot(np.arange(30), metadata = {'longname':'plot name',
                                           'xdata':{'longname': 'xdata name', 'unit': '[s]'},
                                           'ydata':{'longname': 'ydata name', 'unit': '[V]'}})
           plot(np.arange(30), metadata = 'metadata', xmetadata='xdata')

           note: "name" in metadata is reseved, and piScope automatically
                 fills using axis lables.
    '''
    meta0 = kargs.pop('metadata', None)
    if isinstance(meta0, dict):
        meta = OrderedDict()
        for k in meta0:
            if k.endswith('data'):
                if not 'data' in meta:
                    meta['data'] = {}
                meta['data'][k] = OrderedDict(meta0[k])
            else:
                meta[k] = meta0[k]
    else:
        meta = OrderedDict()
        if meta0 is not None:
            meta['name'] = ''
            meta['long_description'] = meta0

        names = []
        for k in kargs:
            if k.endswith('metadata'):
                names.append(k)
                header = k[:-8]
                if header == '':
                    pass
                else:
                    if not 'data' in meta:
                        meta['data'] = {}
                    meta['data'][header+'data'] = OrderedDict()
                    meta['data'][header+'data']['name'] = ''
                    meta['data'][header+'data']['long_description'] = kargs[k]
        if len(names) == 0:
            return None
        for name in names:
            del kargs[name]
    return meta


def allow_interactive_call(method):
    @wraps(method)
    def method2(self, *args, **kargs):
        metadata = pop_metadata(kargs)
        update, kargs = ProcessKeywords(
            kargs, 'update', value=self._interactive_update)
        hold, kargs = ProcessKeywords(kargs, 'hold', value=True)
        autonext, kargs = ProcessKeywords(kargs, 'autonext', value=True)

        f_page = self.get_page()
        if f_page is None:
            self.addpage()
        f_axes = f_page.get_axes(self.isec())
        if not self.canvas.hold_once() and not hold and self.isec() != -1:
            if has_plot(f_axes) and autonext:
                self.increment_isec()
                f_axes = f_page.get_axes(self.isec())
            f_axes.cla()
        if self.canvas.hold_once() and not hold and self.isec() != -1:
            f_axes.cla()

        ret = method(self, *args, **kargs)

        if ret is None:
            return
        if isinstance(ret, tuple):
            obj = ret[-1]
        else:
            obj = ret
        self.canvas.hold_once(False)
        obj = self.add_interative(obj)
        if metadata is not None:
            obj.setvar('metadata', OrderedDict(metadata))

        if update:
            ifigure.events.SendChangedEvent(
                self.book, w=self, useProcessEvent=True)
            ifigure.events.SendPVDrawRequest(
                self.book, wait_idle=True, refresh_hl=False)
        return obj
    import ifigure.interactive
    if hasattr(ifigure.interactive, method.__name__):
        f = getattr(ifigure.interactive, method.__name__)
        method2.__doc__ = f.__doc__
    return method2


def allow_interactive_call_autonext_false(method):
    @wraps(method)
    def method2(self, *args, **kargs):

        update, kargs = ProcessKeywords(
            kargs, 'update', value=self._interactive_update)
        hold, kargs = ProcessKeywords(kargs, 'hold', value=True)
        autonext, kargs = ProcessKeywords(kargs, 'autonext', value=False)

        f_page = self.get_page()
        if f_page is None:
            self.addpage()
        f_page = self.get_page()
        f_axes = f_page.get_axes(self.isec())
        if not self.canvas.hold_once() and not hold and self.isec() != -1:
            if has_plot(f_axes) and autonext:
                self.increment_isec()
                f_axes = f_page.get_axes(self.isec())
            f_axes.cla()
        if self.canvas.hold_once() and not hold and self.isec() != -1:
            f_axes.cla()

        ret = method(self, *args, **kargs)
        if ret is None:
            return
        if isinstance(ret, tuple):
            obj = ret[-1]
        else:
            obj = ret

        self.canvas.hold_once(False)
        obj = self.add_interative(obj)

        if update:
            ifigure.events.SendChangedEvent(
                self.book, w=self, useProcessEvent=True)
            ifigure.events.SendPVDrawRequest(
                self.book, wait_idle=True, refresh_hl=False)
        return obj
    import ifigure.interactive
    if hasattr(ifigure.interactive, method.__name__):
        f = getattr(ifigure.interactive, method.__name__)
        method2.__doc__ = f.__doc__
    return method2


def allow_interactive_call_page(method):
    @wraps(method)
    def method2(self, *args, **kargs):

        update, kargs = ProcessKeywords(
            kargs, 'update', value=self._interactive_update)
        hold, kargs = ProcessKeywords(kargs, 'hold', value=True)

        ret = method(self, *args, **kargs)

        if ret is None:
            return
        if isinstance(ret, tuple):
            obj = ret[-1]
        else:
            obj = ret

        f_page = self.get_page()
        if f_page is None:
            self.addpage()
        f_page = self.get_page()

        obj = self.add_interative(obj, parent=f_page)
        if update:
            ifigure.events.SendChangedEvent(
                self.book, w=self, useProcessEvent=True)
            ifigure.events.SendPVDrawRequest(
                self.book, wait_idle=True, refresh_hl=False)
        return ret
    import ifigure.interactive
    if hasattr(ifigure.interactive, method.__name__):
        f = getattr(ifigure.interactive, method.__name__)
        method2.__doc__ = f.__doc__
    return method2


def allow_interactive_call2(method):
    @wraps(method)
    def method2(self, *args, **kargs):
        from ifigure.utils.cbook import ProcessKeywords
        update, kargs = ProcessKeywords(
            kargs, 'update', value=self._interactive_update)
        hold, kargs = ProcessKeywords(kargs, 'hold', value=True)

        ret = method(self, *args, **kargs)
        if update:
            ifigure.events.SendChangedEvent(
                self.book, w=self, useProcessEvent=True)
            ifigure.events.SendPVDrawRequest(
                self.book, wait_idle=True, refresh_hl=False)
        return ret
    import ifigure.interactive
    if hasattr(ifigure.interactive, method.__name__):
        f = getattr(ifigure.interactive, method.__name__)
        method2.__doc__ = f.__doc__
    return method2


def share_doc_string_simple(method):
    @wraps(method)
    def method2(self, *args, **kargs):
        ret = method(self, *args, **kargs)
        return ret
    import ifigure.interactive
    f = getattr(ifigure.interactive, method.__name__)
    method2.__doc__ = f.__doc__
    return method2


def share_doc_string(method):
    @wraps(method)
    def method2(self, *args, **kargs):
        update, kargs = ProcessKeywords(kargs, 'update',
                                        value=self._interactive_update)
        hold, kargs = ProcessKeywords(kargs, 'hold', value=True)
        ret = method(self, *args, **kargs)
        return ret
    import ifigure.interactive
    f = getattr(ifigure.interactive, method.__name__)
    method2.__doc__ = f.__doc__
    return method2


class BookViewerInteractive(object):
    def __init__(self):
        self._interactive_update = True
        self._interactive_update_request = ([], [])

    def add_interative(self, obj, parent=None, floating=False):
        # first check if any page does exist

        f_page = self.get_page()
        if parent is None:
            parent = self.get_axes(ipage=None, iaxes=self.isec())
        if parent is None:
            raise InteractiveCommandError('no axes exists')

        name = parent.get_next_name(obj.get_namebase())
        if self._interactive_update:
            parent.add_child(name, obj)
        else:
            parent.add_child(name, obj, z_base=self)
        if floating:
            obj._floating = True

        try:
            if self._interactive_update:
                from ifigure.mto.fig_axes import FigAxes
                if isinstance(parent, FigAxes):
                    parent.adjust_axes_range()
                obj.realize()
            else:
                self._interactive_update_request[0].append(obj)
                if not parent in self._interactive_update_request[1]:
                    self._interactive_update_request[1].append(parent)
        except Exception:
            logging.exception("ifigure_app: realize() failed")
            obj.destroy()
            return None
        return obj

    def increment_isec(self):
        '''
        increment_isec moves the current axes to the next
        position by incrementing isec. if isec+1=nsec,
        isec is set to zero
        '''
        f_page = self.get_page()
        f_axes = f_page.get_axes(self.isec()+1)
        if f_axes is None:
            self.canvas._isec = 0
        else:
            self.canvas._isec = self.isec()+1

    def decrement_isec(self):
        '''
        increment_isec moves the current axes to the next
        position by incrementing isec. if isec+1=nsec,
        isec is set to zero
        '''
        f_page = self.get_page()
        f_axes = f_page.get_axes(self.isec()-1)
        if f_axes is None:
            self.canvas._isec = f_page.num_axes(include_floating=True)-1
        else:
            self.canvas._isec = self.isec()-1

    def update(self, value):
        from ifigure.mto.fig_axes import FigAxes
        if value:
            if not self._interactive_update:
                for parent in self._interactive_update_request[1]:
                    if isinstance(parent, FigAxes):
                        parent.adjust_axes_range()
                for obj in self._interactive_update_request[0]:
                    try:
                        obj.realize()
                    except:
                        traceback.print_exc()
                        print("failed to realize (destroied)" + str(obj))
                        obj.destroy()
                self._interactive_update_request = ([], [])
                ifigure.events.SendChangedEvent(
                    self.book, w=self, useProcessEvent=True)
                ifigure.events.SendPVDrawRequest(
                    self.book, wait_idle=True, refresh_hl=False)
        else:
            f_page = self.get_page()
            if f_page is not None:
                self._z_base = max([obj2.getp('zorder')
                                    for obj2 in f_page.walk_figobj()])
            else:
                self._z_base = 1
        self._interactive_update = value

    @share_doc_string
    def addpage(self, num=1, before=False):
        for k in range(num):
            page = self.new_page()
            i = self.add_page(page, before=before)

        fig_p = self.get_page(ipage=i)
        fig_b = self.book
        ifigure.events.SendChangedEvent(fig_b, w=self, useProcessEvent=True)
        ifigure.events.SendShowPageEvent(fig_p, w=self)
        return fig_p

    @share_doc_string
    def showpage(self, ipage):
        self.show_page(ipage)

    @share_doc_string
    def delpage(self):
        ipage = self.ipage
        fig_b = self.get_page(ipage=ipage).get_figbook()
        if fig_b.num_page() == 1:
            raise InteractiveCommandError('can not delete the current page')
        self.del_page()
        ifigure.events.SendChangedEvent(fig_b, w=self, useProcessEvent=True)

    @allow_interactive_call2
    def cla(self, reset_color_cycle=True):
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError()

        axes = self.get_axes(ipage=None, iaxes=self.isec())
        axes.cla()
        axes.reset_color_cycle()

    @allow_interactive_call2
    def cls(self, obj=None):
        self.clf(obj=obj)

    @allow_interactive_call2
    def clf(self, obj=None):
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError('no page exists')
        for axes in fig_p.walk_axes():
            axes.reset_color_cycle()
            children = [child for child in axes.walk_tree()]
            if obj is not None:
                children = [obj] if obj in children else []
            for child in children:
                if (not hasattr(child, '_generic_axes') or
                        not child._generic_axes):
                    child.destroy()
                else:
                    child.reset_color_cycle()
        if self.nsec() > 0:
            self.isec(0)

    def _do_subplot(self, *args, **kargs):
        fig_p = self.get_page()
        if fig_p is None:
            raise NoPageError('no page exists')

        if len(args) == 0:
            return fig_p.num_axes()

        self.set_section(*args, **kargs)
        ax = self.get_axes(ipage=None, iaxes=self.isec())
        if ax is None:
            ax = self.get_axes(ipage=None, iaxes=0)
        return ax

    @allow_interactive_call2
    def nsec(self, *args, **kargs):
        return self._do_subplot(*args, **kargs)

    @allow_interactive_call2
    def subplot(self, *args, **kargs):
        return self._do_subplot(*args, **kargs)

    @share_doc_string
    def isec(self, i=None):
        '''
        isce/isection control current axes.
        '''
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError('no page exists')

        if i is not None:

            fig_p = self.get_page(ipage=None)
            if fig_p.get_axes(iax=i) is None:
                raise InteractiveCommandError('axes['+str(fig_p.num_axes()) +
                                              '] does not exist')
            self.canvas._isec = i
            return self.canvas._isec
        else:
            if self.canvas._isec == -1:
                fig_ax = self.get_axes(ipage=None, iaxes=0)
                if fig_ax is not None:
                    self.set_axes(fig_ax)
                    self.canvas._isec = 0
            return self.canvas._isec

    @allow_interactive_call2
    def suptitle(self, txt, size=None, color=None):
        fig_page = self.get_page(ipage=None)
        fig_page.set_suptitle(txt, a=None, size=size, color=color)

    @allow_interactive_call2
    def title(self, txt, size=None, color=None):
        '''
        set title of current section
        '''
        fig_ax = self.get_axes(ipage=None, iaxes=self.isec())
        if fig_ax is None:
            return
        if len(fig_ax._artists) == 0:
            fig_ax.getp('title_labelinfo')[0] = txt
        else:
            v = fig_ax.get_title(fig_ax._artists[0])
            v[0] = txt
            if size is not None:
                v[-1] = 'default' if size == 'default' else float(size)
            if color is not None:
                v[1] = color
            fig_ax.set_title(v, fig_ax._artists[0])
        fig_ax.set_bmp_update(False)

    def _set_label(self, txt, name, size, color):
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError("no page exists")
        fig_ax = self.get_axes(ipage=None, iaxes=self.isec())
        if fig_ax.get_axis_param(name) is None:
            raise InteractiveCommandError('Axis ('+name + ') does not exist')

        v = fig_ax.get_axlabel(fig_ax._artists[0], name)
        if txt is None and size is None:
            return v
        if txt is not None:
            v[0] = txt
        if size is not None:
            v[-1] = 'default' if size == 'default' else float(size)
        if color is not None:
            v[1] = color
        fig_ax.set_axlabel((name, v), fig_ax._artists[0])
        fig_ax.set_bmp_update(False)
        return txt

    @allow_interactive_call2
    def xlabel(self, txt=None, name='x', size=None, color=None):
        return self._set_label(txt, name, size, color)

    @allow_interactive_call2
    def xtitle(self, txt=None, name='x', size=None, color=None):
        return self._set_label(txt, name, size, color)

    @allow_interactive_call2
    def ylabel(self, txt=None, name='y', size=None, color=None):
        return self._set_label(txt, name, size, color)

    @allow_interactive_call2
    def ytitle(self, txt=None, name='y', size=None, color=None):
        return self._set_label(txt, name, size, color)

    @allow_interactive_call2
    def zlabel(self, txt=None, name='z', size=None, color=None):
        return self._set_label(txt, name, size, color)

    @allow_interactive_call2
    def ztitle(self, txt=None, name='z', size=None, color=None):
        return self._set_label(txt, name, size, color)

    @allow_interactive_call2
    def clabel(self, txt=None, name='c', size=None, color=None):
        return self._set_label(txt, name, size, color)

    @allow_interactive_call2
    def ctitle(self, txt=None, name='c', size=None, color=None):
        return self._set_label(txt, name, size, color)

    def _set_log(self, v, name, base=None, linthresh=None, linscale=None):
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError("no page exists")
        fig_ax = self.get_axes(ipage=None, iaxes=self.isec())
        if fig_ax.get_axis_param(name) is None:
            raise InteractiveCommandError('Axis ('+name + ') does not exist')

        val = fig_ax.get_axrangeparam(fig_ax._artists[0], name)
        if v == 'symlog':
            val[3] = 'symlog'
        elif v == False:
            val[3] = 'linear'
        elif v == True:
            val[3] = 'log'
        if base is not None:
            val[0] = str(base)
        if linthresh is not None:
            val[4] = linthresh
        if linscale is not None:
            val[5] = linscale
        request = [(name, val)]
        request = fig_ax.compute_new_range(request=[(name, val)])
        fig_ax.perform_range_request(request)

    @allow_interactive_call2
    def xlog(self, value=True, base=None):
        self._set_log(value, 'x', base=base)

    @allow_interactive_call2
    def ylog(self, value=True, base=None):
        self._set_log(value, 'y', base=base)

    @allow_interactive_call2
    def clog(self, value=True, base=None):
        self._set_log(value, 'c', base=base)

    @allow_interactive_call2
    def zlog(self, value=True, base=None):
        self._set_log(value, 'z', base=base)

    @allow_interactive_call2
    def xlinear(self, value=False):
        self._set_log(value, 'x')

    @allow_interactive_call2
    def ylinear(self, value=False):
        self._set_log(value, 'y')

    @allow_interactive_call2
    def clinear(self, value=False):
        self._set_log(value, 'c')

    @allow_interactive_call2
    def zlinear(self, value=False):
        self._set_log(value, 'z')

    @allow_interactive_call2
    def xsymlog(self, base=None, linthresh=None, linscale=None,
                name='x'):
        self._set_log('symlog', name, base=base, linthresh=linthresh,
                      linscale=linscale)

    @allow_interactive_call2
    def ysymlog(self, base=None, linthresh=None, linscale=None,
                name='y'):
        self._set_log('symlog', name, base=base, linthresh=linthresh,
                      linscale=linscale)

    @allow_interactive_call2
    def zsymlog(self, base=None, linthresh=None, linscale=None,
                name='z'):
        self._set_log('symlog', name, base=base, linthresh=linthresh,
                      linscale=linscale)

    @allow_interactive_call2
    def csymlog(self, base=None, linthresh=None, linscale=None,
                name='c'):
        self._set_log('symlog', name, base=base, linthresh=linthresh,
                      linscale=linscale)

    def _set_auto(self, name):
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError("no page exists")
        axes = self.get_axes(ipage=None, iaxes=self.isec())
        if axes.get_axis_param(name) is None:
            raise InteractiveCommandError('Axis ('+name + ') does not exist')

        v = axes.get_axrangeparam(axes._artists[0], name)
        v[1] = True
        request = axes.compute_new_range(request=[(name, v)])
        for r in request:
            axes.set_axrangeparam(r, axes._artists[0])
        ifigure.events.SendPVDrawRequest(axes, w=self, wait_idle=False,
                                         refresh_hl=False)

    @allow_interactive_call2
    def xauto(self, name='x'):
        self._set_auto(name)

    @allow_interactive_call2
    def yauto(self, name='y'):
        self._set_auto(name)

    @allow_interactive_call2
    def zauto(self, name='z'):
        self._set_auto(name)

    @allow_interactive_call2
    def cauto(self, name='c'):
        self._set_auto(name)

    def _set_lim(self, *range, **kargs):
        if len(range) == 2:
            val = (range[0], range[1])
        elif len(range) == 0:
            val = None
        else:
            val = range[0]

        name = kargs['name']

        if (name.startswith('x') or name.startswith('y') or
                name.startswith('z')):
            axis_param_names0 = ('tick_position', 'ticks', 'tcolor',
                                 'lcolor', 'lsize', 'otcolor', 'otsize')
            axis_param_names = ('tposition', 'ticks', 'tcolor',
                                'color', 'size', 'ocolor', 'osize')
            axis_param = {x: kargs[x] for x in axis_param_names if x in kargs}

        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError("no page exists")
        axes = self.get_axes(ipage=None, iaxes=self.isec())
        if axes.get_axis_param(name) is None:
            raise InteractiveCommandError('Axis ('+name + ') does not exist')
        v = axes.get_axrangeparam(axes._artists[0], name)
        if val is None:
            return v[2]
        v[1] = False
        v[2] = val
        axes.set_axrangeparam((name, v), axes._artists[0])

        if (name.startswith('x') or name.startswith('y') or
                name.startswith('z')):

            ap = axes.get_axis_param(name)
            for x, y in zip(axis_param_names0, axis_param_names):
                if y in axis_param:
                    setattr(ap, x, axis_param[y])
            artist = axes.get_axes_artist_by_name(name)[0]
            ap.set_tickparam(artist, axes)
            ap.set_ticks(artist)

        axes.set_bmp_update(False)

    @allow_interactive_call2
    def xlim(self, *range, **kargs):
        name, kargs = ProcessKeywords(kargs, 'name', value='x')
        kargs['name'] = name
        return self._set_lim(*range, **kargs)

    @allow_interactive_call2
    def ylim(self, *range, **kargs):
        name, kargs = ProcessKeywords(kargs, 'name', value='y')
        kargs['name'] = name
        return self._set_lim(*range, **kargs)

    @allow_interactive_call2
    def zlim(self, *range, **kargs):
        name, kargs = ProcessKeywords(kargs, 'name', value='z')
        kargs['name'] = name
        return self._set_lim(*range, **kargs)

    @allow_interactive_call2
    def clim(self, *range, **kargs):
        name, kargs = ProcessKeywords(kargs, 'name', value='c')
        kargs['name'] = name
        return self._set_lim(*range, **kargs)

    @allow_interactive_call2
    def twinx(self):
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError('no page exists')
        axes = self.get_axes(ipage=None, iaxes=self.isec())
        if axes.hastwin('x'):
            raise InteractiveCommandError('axes has twin already')
        #cvalue = axes.get_twinx(None)
        axes.twinx()
        value = [p.make_save_data(axes) for p in axes._yaxis]
        value.append(axes._yaxis[-1].make_save_data(axes))
        value[-1]["_ax_idx"] = []
        value = (True, value)
        axes.del_twinx()
        axes.set_twinx(value, None)
        axes.set_bmp_update(False)

    @allow_interactive_call2
    def twiny(self):
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError('no page exists')
        axes = self.get_axes(ipage=None, iaxes=self.isec())
        if axes.hastwin('y'):
            raise InteractiveCommandError('axes has twin already')
        #cvalue = axes.get_twiny(None)
        axes.twiny()
        value = [p.make_save_data(axes) for p in axes._xaxis]
        value.append(axes._xaxis[-1].make_save_data(axes))
        value[-1]["_ax_idx"] = []
        value = (True, value)
        axes.del_twiny()
        axes.set_twiny(value, None)
        axes.set_bmp_update(False)

    @allow_interactive_call2
    def threed(self, *args):
        import ifigure.utils.cbook as cbook
        ax = self.get_axes(ipage=None, iaxes=self.isec())
        if ax is None:
            return
        if len(args) == 0:
            return ax.get_3d()
        v = cbook.on_off_args(args[0])
        ax.set_3d(v)
        # ax.del_artist(delall=True)
        # ax.realize()
        self.draw()
        self.canvas.set_axes_selection(ax._artists[0])
        self.canvas.toolbar.Show3DMenu()

    def oplot(self, *args, **kargs):
        kargs['hold'] = True
        v = self.plot(*args, **kargs)
        return v

    def oerrorbar(self, *args, **kargs):
        kargs['hold'] = True
        v = self.errorbar(*args, **kargs)
        return v

    @allow_interactive_call
    def timetrace(self, *args, **kargs):
        from ifigure.mto.fig_plot import TimeTrace
        obj = self._plot(*args, **kargs)
        if obj is not None:
            obj.__class__ = TimeTrace
            obj._use_decimate = True
        return obj

    @allow_interactive_call
    def plot(self, *args, **kargs):
        return self._plot(*args, **kargs)

    @allow_interactive_call
    def loglog(self, *args, **kargs):
        ret = self._plot(*args, **kargs)
        self._set_log(True, 'x')
        self._set_log(True, 'y')
        return ret

    @allow_interactive_call
    def semilogy(self, *args, **kargs):
        ret = self._plot(*args, **kargs)
        self._set_log(True,  'y')
        self._set_log(False, 'x')
        return ret

    @allow_interactive_call
    def semilogx(self, *args, **kargs):
        ret = self._plot(*args, **kargs)
        self._set_log(True,   'x')
        self._set_log(False,  'y')
        return ret

    @allow_interactive_call
    def plotc(self, *args, **kargs):
        from ifigure.mto.fig_plotc import FigPlotC
        kargs['cls'] = FigPlotC
        return self._plot(*args, **kargs)

    @allow_interactive_call
    def errorbar(self, *args, **kargs):
        from ifigure.mto.fig_plot import FigPlot
        kargs['mpl_command'] = 'errorbar'
        return self._plot(*args, **kargs)

    @allow_interactive_call
    def errorbarc(self, *args, **kargs):
        from ifigure.mto.fig_plotc import FigPlotC
        kargs['cls'] = FigPlotC
        kargs['mpl_command'] = 'errorbar'
        return self._plot(*args, **kargs)

    def _plot(self, *args, **kargs):
        from ifigure.utils.cbook import isiterable, isndarray, isdynamic, issequence, isnumber

        def convert_2_real_array(a):
            '''
            it actually keep complex as complex (2015.09.29)
            '''
            if a is None:
                return None
            if not isiterable(a):
                return a
            if len(a) == 1:
                return a
            if np.array(a).dtype in _dtype_numbers:
                return a
            else:
                x = np.array(a).astype(float)
                return x

        from ifigure.mto.fig_plot import FigPlot
        from ifigure.mto.fig_grp import FigGrp

        cls, kargs = ProcessKeywords(kargs, 'cls', value=FigPlot)
        # this is to analyze arguments
        try:
            obj = cls(*args, **kargs)
        except ValueError as x:
            traceback.print_exc()
            print(x)
            return

        x = convert_2_real_array(obj.getvar('x'))
        y = convert_2_real_array(obj.getvar('y'))
        z = convert_2_real_array(obj.getvar('z'))
        c = convert_2_real_array(obj.getvar('c'))
        xerr = convert_2_real_array(obj.getvar('xerr'))
        yerr = convert_2_real_array(obj.getvar('yerr'))

        if (isinstance(y, np.ndarray) and
                x.ndim == 1 and y.ndim == 2):
            s = obj.getvar('s')
            obj = FigGrp()
            for k in range(y.shape[0]):
                if 'xerr' in kargs:
                    if isiterable(xerr):
                        kargs['xerr'] = xerr[k]
                if 'yerr' in kargs:
                    if isiterable(yerr):
                        kargs['yerr'] = yerr[k]
                if c is not None:
                    kargs['c'] = c[k]
                if z is None:
                    obj2 = FigPlot(x, y[k], s, **kargs)
                else:
                    obj2 = FigPlot(x, y[k], z[k], s, **kargs)
                name = obj.get_next_name(obj2.get_namebase())
                obj.add_child(name, obj2)
        elif (isinstance(x, np.ndarray) and
              x.ndim == 2 and y.ndim == 1):
            s = obj.getvar('s')
            obj = FigGrp()
            for k in range(x.shape[0]):
                if 'xerr' in kargs:
                    if isiterable(xerr):
                        kargs['xerr'] = xerr[k]
                if 'yerr' in kargs:
                    if isiterable(yerr):
                        kargs['yerr'] = yerr[k]
                if c is not None:
                    kargs['c'] = c[k]
                if z is None:
                    obj2 = FigPlot(x[k], y, s, **kargs)
                else:
                    obj2 = FigPlot(x[k], y, z[k], s, **kargs)
                name = obj.get_next_name(obj2.get_namebase())
                obj.add_child(name, obj2)
        elif (isinstance(x, np.ndarray) and
              x.ndim == 2 and y.ndim == 2):
            s = obj.getvar('s')
            obj = FigGrp()
            for k in range(x.shape[0]):
                if 'xerr' in kargs:
                    if isiterable(xerr):
                        kargs['xerr'] = xerr[k]
                if 'yerr' in kargs:
                    if isiterable(yerr):
                        kargs['yerr'] = yerr[k]
                if c is not None:
                    kargs['c'] = c[k]
                if z is None:
                    obj2 = FigPlot(x[k], y[k], s, **kargs)
                else:
                    obj2 = FigPlot(x[k], y[k], z[k], s, **kargs)
                name = obj.get_next_name(obj2.get_namebase())
                obj.add_child(name, obj2)
        else:
            obj.setvar('x', x)
            obj.setvar('y', y)
            obj.setvar('z', z)
            obj.setvar('xerr', xerr)
            obj.setvar('yerr', yerr)

        return obj

    @allow_interactive_call
    def triplot(self, *args, **kargs):
        from ifigure.mto.fig_triplot import FigTriplot
        try:
            obj = FigTriplot(*args, **kargs)
        except ValueError as x:
            print(x)
            return

        return obj

    @allow_interactive_call
    def annotate(self, *args, **kargs):
        from ifigure.mto.fig_annotate import FigAnnotate
        obj = FigAnnotate(*args, **kargs)
        return obj

    @allow_interactive_call
    def ispline(self, *args, **kargs):
        from ifigure.mto.fig_spline import FigSpline
        try:
            obj = FigSpline(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def contour(self, *args, **kargs):
        from ifigure.mto.fig_contour import FigContour
        if not "FillMode" in kargs:
            kargs["FillMode"] = False
        try:
            obj = FigContour(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def contourf(self, *args, **kargs):
        from ifigure.mto.fig_contour import FigContour
        kargs["FillMode"] = True
        try:
            obj = FigContour(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def scatter(self, *args, **kargs):
        from ifigure.mto.fig_scatter import FigScatter
        try:
            obj = FigScatter(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def hist(self, x, *args, **kargs):
        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        axes = fig_axes._artists[0]

        color = kargs.pop('color', None)
        weights = kargs.pop('weights', None)

        # Process 'x'
        # This part is copied from matplotlib._axes.
        from matplotlib.cbook import iterable
        if isinstance(x, np.ndarray) or not iterable(x[0]):
            # TODO: support masked arrays;
            x = np.asarray(x)
            if x.ndim == 2:
                x = x.T  # 2-D input with columns as datasets; switch to rows
            elif x.ndim == 1:
                x = x.reshape(1, x.shape[0])  # new view, single row
            else:
                raise ValueError("x must be 1D or 2D")
            if x.shape[1] < x.shape[0]:
                warnings.warn(
                    '2D hist input should be nsamples x nvariables;\n '
                    'this looks transposed (shape is %d x %d)' % x.shape[::-1])
        else:
            # multiple hist with data of different length
            x = [np.asarray(xi) for xi in x]

        nx = len(x)  # number of dataset
        from ifigure.matplotlib_mod.mpl_utils import get_color_cycle_list
        if color is None:
            color = get_color_cycle_list(axes)
        else:
            color = mcolors.colorConverter.to_rgba_array(color)
            if len(color) != nx:
                raise ValueError("color kwarg must have one color per dataset")

        # We need to do to 'weights' what was done to 'x'
        if weights is not None:
            if isinstance(weights, np.ndarray) or not iterable(weights[0]):
                w = np.array(weights)
                if w.ndim == 2:
                    w = w.T
                elif w.ndim == 1:
                    w.shape = (1, w.shape[0])
                else:
                    raise ValueError("weights must be 1D or 2D")
            else:
                w = [np.asarray(wi) for wi in weights]

            if len(w) != nx:
                raise ValueError('weights should have the same shape as x')
            for i in range(nx):
                if len(w[i]) != len(x[i]):
                    raise ValueError(
                        'weights should have the same shape as x')
        else:
            w = [None]*nx

        #
        #  end of processing
        from ifigure.mto.fig_hist import FigHist
        from ifigure.mto.fig_grp import FigGrp
        import wx
        try:
            if nx == 1:
                kargs['weights'] = w[0]
                kargs['color'] = color[0]
                obj = FigHist(x[0], *args, **kargs)
                wx.CallAfter(obj.call_adjust_range)
            else:
                obj = FigGrp()
                for k, xx in enumerate(x):
                    kargs['weights'] = w[k]
                    kargs['color'] = color[k]
                    obj2 = FigHist(xx, *args, **kargs)
                    name = obj.get_next_name(obj2.get_namebase())
                    obj.add_child(name, obj2)
                wx.CallAfter(obj2.call_adjust_range)
        except ValueError as x:
            print(x)
            return

        return obj

    @allow_interactive_call
    def tricontour(self, *args, **kargs):
        from ifigure.mto.fig_tricontour import FigTricontour
        if not "FillMode" in kargs:
            kargs["FillMode"] = False
        try:
            obj = FigTricontour(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def tricontourf(self, *args, **kargs):
        from ifigure.mto.fig_tricontour import FigTricontour
        kargs["FillMode"] = True
        try:
            obj = FigTricontour(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def image(self, *args, **kargs):
        if 'use_tri' in kargs and kargs['use_tri']:
            return tripcolor(*args, **kargs)

        from ifigure.mto.fig_image import FigImage
        try:
            obj = FigImage(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def quiver(self, *args, **kargs):
        from ifigure.mto.fig_quiver import FigQuiver, FigQuiver3D

        is3D = self.threed()
        try:
            if is3D:
                obj = FigQuiver3D(*args, **kargs)
            else:
                obj = FigQuiver(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def quiver3d(self, *args, **kargs):
        from ifigure.mto.fig_quiver import FigQuiver, FigQuiver3D

        self.threed('on')
        try:
            obj = FigQuiver3D(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    import matplotlib.mlab as mlab
#    @allow_interactive_call

    def specgram(self, x, NFFT=256,
                 Fs=2,
                 Fc=0,
                 detrend=mlab.detrend_none,
                 window=mlab.window_hanning,
                 noverlap=128,
                 xextent=None,
                 pad_to=None,
                 sides='default',
                 scale_by_freq=None,
                 **kwargs):
        '''
        plot spectrogram. Run matplotlib.pyplot.specgram
        and call image using the returnd spectrum.
        keywords are the same as specgram.
        '''
        fig_p = self.get_page(ipage=None)
        if fig_p is None:
            raise NoPageError('no page exists')

#        axes = self.get_axes(ipage=None, iaxes=self.isec())
        Pxx, freqs, bins = mlab.specgram(x, NFFT=NFFT,
                                         Fs=Fs,
                                         detrend=detrend,
                                         window=window,
                                         noverlap=noverlap,
                                         pad_to=None,
                                         sides='default',
                                         scale_by_freq=None)
#        im.remove()
        freqs = freqs + Fc
        obj = self.image(bins, freqs, Pxx, **kwargs)
        return Pxx, freqs, bins, obj

    @allow_interactive_call
    def spec(self, *args, **kargs):
        from ifigure.mto.fig_spec import FigSpec
        try:
            obj = FigSpec(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def tripcolor(self, *args, **kargs):
        from ifigure.mto.fig_tripcolor import FigTripcolor
        try:
            obj = FigTripcolor(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def axline(self, *args, **kargs):
        from ifigure.mto.fig_axline import FigAxline
        try:
            obj = FigAxline(*args, **kargs)
        except ValueError as x:
            print(x)
            return

        return obj

    @allow_interactive_call
    def axlinec(self, *args, **kargs):
        from ifigure.mto.fig_axlinec import FigAxlineC
        try:
            obj = FigAxlineC(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call
    def axspan(self, *args, **kargs):
        from ifigure.mto.fig_axspan import FigAxspan
        try:
            obj = FigAxspan(*args, **kargs)
        except ValueError as x:
            print(x)
            return

        return obj

    @allow_interactive_call
    def axspanc(self, *args, **kargs):
        from ifigure.mto.fig_axspanc import FigAxspanC
        try:
            obj = FigAxspanC(*args, **kargs)
        except ValueError as x:
            print(x)
            return

        return obj

    @allow_interactive_call_autonext_false
    def text(self, *args, **kargs):
        from ifigure.mto.fig_text import FigText

        try:
            if not 'trans' in kargs:
                kargs['trans'] = ['axes']*2
            else:
                if isinstance(kargs['trans'], str):
                    kargs['trans'] = [kargs['trans']]*2

            obj = FigText(*args, **kargs)
        except ValueError as x:
            print(x)
            return

        return obj

    @allow_interactive_call_autonext_false
    def arrow(self, *args, **kargs):
        from ifigure.mto.fig_arrow import FigArrow

        try:
            if not 'trans' in kargs:
                kargs['trans'] = ['axes']*4
            else:
                if isinstance(kargs['trans'], str):
                    kargs['trans'] = [kargs['trans']]*4
            #kargs['autonext'] = False
            obj = FigArrow(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call_page
    def figarrow(self, *args, **kargs):
        from ifigure.mto.fig_arrow import FigArrow

        try:
            if not 'trans' in kargs:
                kargs['trans'] = ['figure']*4
            obj = FigArrow(*args, **kargs)
        except ValueError as x:
            print(x)
            return
        return obj

    @allow_interactive_call_page
    def figtext(self, *args, **kargs):
        from ifigure.mto.fig_text import FigText

        obj = FigText(*args, **kargs)
        if obj is None:
            return

        return obj

    @allow_interactive_call_autonext_false
    def legend(self, *args, **kargs):
        from ifigure.mto.fig_legend import FigLegend
        if len(args) == 0:
            return
        if len(args) == 1 and isinstance(args[0], str):
            args = ((args[0],),)
        if "axes2" in kargs:
            kargs["container_idx"] = 1 if kargs["axes2"] else 0
            del kargs["axes2"]

        f_page = self.get_page()
        if f_page is None:
            return None
        f_axes = f_page.get_axes(self.isec())
        for name, obj in f_axes.get_children():
            if isinstance(obj, FigLegend):
                labels = obj.getvar('legendlabel')
                for arg in args:
                    labels = sum((labels, arg), ())
                obj.setvar('legendlabel', labels)
                obj.setp('legendlabel', labels)
                obj.refresh_artist()
                ifigure.events.SendChangedEvent(
                    self.book, w=self, useProcessEvent=True)
                ifigure.events.SendPVDrawRequest(
                    self.book, wait_idle=True, refresh_hl=False)
                return
        #kargs['autonext'] = False
        obj = FigLegend(*args, **kargs)
        return obj

    @allow_interactive_call
    def fill(self, *args, **kargs):
        from ifigure.mto.fig_fill import FigFill
        kargs['mpl_command'] = 'fill'
        obj = FigFill(*args, **kargs)
        return obj

    @allow_interactive_call
    def fill_between(self, *args, **kargs):
        from ifigure.mto.fig_fill import FigFill
        kargs['mpl_command'] = 'fill_between'
        obj = FigFill(*args, **kargs)
        return obj

    @allow_interactive_call
    def fill_betweenx(self, *args, **kargs):
        from ifigure.mto.fig_fill import FigFill
        kargs['mpl_command'] = 'fill_betweenx'
        obj = FigFill(*args, **kargs)
        return obj

    @allow_interactive_call
    def fill_between_3d(self, x1, y1, z1, x2, y2, z2, c='b', w=0, **kargs):
        from ifigure.mto.fig_solid import FigSolid

        verts1 = np.vstack([np.array(x1, copy=False),
                            np.array(y1, copy=False),
                            np.array(z1, copy=False), ]).transpose()
        verts2 = np.vstack([np.array(x2, copy=False),
                            np.array(y2, copy=False),
                            np.array(z2, copy=False), ]).transpose()

        x = np.arange(len(verts1))
        y = np.arange(2)
        X, Y = np.meshgrid(x, y)

        from ifigure.utils.triangulation_wrapper import delaunay

        idxset = delaunay(X.flatten(), Y.flatten())
        verts = np.vstack([verts1, verts2])

        args = (verts, idxset)

        kargs['linewidth'] = w
        kargs['facecolor'] = c

        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        if not fig_axes.get_3d():
            fig_axes.set_3d(True)

        obj = FigSolid(*args, **kargs)
        return obj

    @allow_interactive_call
    def surf(self, *args, **kargs):
        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        if not fig_axes.get_3d():
            return

        from ifigure.mto.fig_surface import FigSurface
        obj = FigSurface(*args, **kargs)
        return obj

    @allow_interactive_call
    def surface(self, x, y, z, **kargs):
        return self.surf(x, y, z, **kargs)

    @allow_interactive_call
    def revolve(self, *args, **kargs):
        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        if not fig_axes.get_3d():
            return

        from ifigure.mto.fig_surface import FigRevolve
        obj = FigRevolve(*args, **kargs)
        return obj

    @allow_interactive_call
    def trisurf(self, *args, **kargs):
        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        if not fig_axes.get_3d():
            return

        from ifigure.mto.fig_trisurface import FigTrisurface
        obj = FigTrisurface(*args, **kargs)
        return obj

    @allow_interactive_call
    def solid(self, *args, **kargs):
        from ifigure.mto.fig_solid import FigSolid
        if args[0].shape[-1] == 4:
            kargs['cz'] = True
            kargs['cdata'] = args[0][..., -1]
            args = (args[0][..., :-1],)

        def_width = 0.0
        if len(args) == 2:
            verts = args[0]
            idxset = args[1]
            if verts.shape[-1] == 4:
                kargs['cz'] = True
                kargs['cdata'] = verts[..., -1]
                verts = verts[..., :-1]
            elif verts.shape[-1] == 2:
                zvalue = kargs.pop('zvalue', 0.0)
                verts = np.hstack((args[0], np.zeros((args[0].shape[0], 1))
                                   + zvalue))
            args = (verts, idxset)
            if idxset.shape[1] == 2:
                def_width = 1.0
        elif len(args) == 1:
            verts = args[0]
            if verts.shape[-1] == 4:
                kargs['cz'] = True
                kargs['cdata'] = verts[..., -1]
                verts = verts[..., :-1]
            elif args[0].shape[-1] == 2:
                zvalue = kargs.pop('zvalue', 0.0)
                verts = np.dstack((args[0], np.zeros((args[0].shape[0], args[0].shape[1], 1))
                                   + zvalue))
            args = (verts,)
            if verts.shape[1] == 2:
                def_width = 1.0
        else:
            assert False, "wrong number of arguments, solid(v, idx) or solid(v)"

        linewidth = kargs.pop("linewidth", def_width)
        kargs['linewidth'] = linewidth
        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        if not fig_axes.get_3d():
            fig_axes.set_3d(True)

        obj = FigSolid(*args, **kargs)
        return obj

    @allow_interactive_call2
    def lighting(self, **kargs):
        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        if not fig_axes.get_3d():
            return
        ax = fig_axes._artists[0]
        l = ax.get_lighting()
        if len(kargs) == 0:
            return l
        for k in kargs:
            if not k in l:
                raise KeyError(str(k) + ' does not exists')
        for k in kargs:
            l[k] = kargs[k]
        fig_axes.set_bmp_update(False)
        self.draw()

    @allow_interactive_call2
    def _view(self, *args, **kargs):
        fig_axes = self.get_axes(ipage=None, iaxes=self.isec())
        if fig_axes is None:
            return
        if not fig_axes.get_3d():
            return
        ax = fig_axes._artists[0]
        if len(args) == 0 and len(kargs) == 0:
            return ax.elev, ax.azim, ax._upvec
        elif len(args) == 3:
            ax.elev = args[0]
            ax.azim = args[1]
            ax._upvec = args[2]
            fig_axes.set_bmp_update(False)
        elif len(args) == 1:
            if args[0] == 'xy':
                if ax.elev == 90 and ax.azim == 180:
                    ax.elev, ax.azim = -90, 180
                else:
                    ax.elev, ax.azim = 90, 180
                ax._upvec = np.array([0, 1, 0])
                fig_axes.set_bmp_update(False)
            elif args[0] == 'xz':
                if ax.elev == 0 and ax.azim == -90:
                    ax.elev, ax.azim = 180, -90
                else:
                    ax.elev, ax.azim = 0, -90
                ax._upvec = np.array([0, 0, 1])
                fig_axes.set_bmp_update(False)
            elif args[0] == 'yx':
                ax.elev, ax.azim = -90, 0
                ax._upvec = np.array([0, 0, 1])
                fig_axes.set_bmp_update(False)
            elif args[0] == 'yz':
                if ax.elev == 0 and ax.azim == 0:
                    ax.elev, ax.azim = 180, 0
                else:
                    ax.elev, ax.azim = 0, 0
                ax._upvec = np.array([0, 0, 1])
                fig_axes.set_bmp_update(False)
            elif args[0] == 'zx':
                ax.elev, ax.azim = 0, 90
                ax._upvec = np.array([1, 0, 0])
                fig_axes.set_bmp_update(False)
            elif args[0] == 'zy':
                ax.elev, ax.azim = 0, -180
                ax._upvec = np.array([0, 1, 0])
                fig_axes.set_bmp_update(False)
            elif args[0] == 'default':
                ax.elev, ax.azim = 30, -60
                ax._upvec = np.array([0, 0, 1])
                fig_axes.set_bmp_update(False)
            elif args[0] == '90':
                ax.rotate_view_90deg(True)
                fig_axes.set_bmp_update(False)
            elif args[0] == '-90':
                ax.rotate_view_90deg(False)
                fig_axes.set_bmp_update(False)
            elif args[0] == 'updown':
                ax._upvec = -ax._upvec
                fig_axes.set_bmp_update(False)
            elif args[0] == 'ortho':
                ax._use_frustum = False
                fig_axes.set_bmp_update(False)
            elif args[0] == 'frustum':
                ax._use_frustum = True
                fig_axes.set_bmp_update(False)
            elif args[0] == 'noclip':
                ax._use_clip = ax._use_clip & 2
                fig_axes.set_bmp_update(False)
            elif args[0] == 'clip':
                ax._use_clip = ax._use_clip | 1
                fig_axes.set_bmp_update(False)
            elif args[0] == 'nocp':
                ax._use_clip = ax._use_clip & 1
                fig_axes.set_bmp_update(False)
            elif args[0] == 'cp':
                ax._use_clip = ax._use_clip | 2
                fig_axes.set_bmp_update(False)

            elif args[0] == 'auto':
                self.xauto()
                self.yauto()
                self.zauto()
                fig_axes.set_aspect('auto')
                fig_axes.set_bmp_update(False)

            elif args[0] == 'equal':
                xlim = self.xlim()
                ylim = self.ylim()
                zlim = self.zlim()
                dx = abs(xlim[1]-xlim[0])
                dy = abs(ylim[1]-ylim[0])
                dz = abs(zlim[1]-zlim[0])
                dd = max((dx, dy, dz))
                self.xlim((xlim[1]+xlim[0]-dd)/2,
                          (xlim[1]+xlim[0]+dd)/2)
                self.ylim((ylim[1]+ylim[0]-dd)/2,
                          (ylim[1]+ylim[0]+dd)/2)
                self.zlim((zlim[1]+zlim[0]-dd)/2,
                          (zlim[1]+zlim[0]+dd)/2)
                fig_axes.set_aspect('equal')
                fig_axes.set_bmp_update(False)
            elif args[0] == 'axesicon':
                ax._show_3d_axes = True
                fig_axes.set_bmp_update(False)
            elif args[0] == 'noaxesicon':
                ax._show_3d_axes = False
                fig_axes.set_bmp_update(False)
            else:
                print('Unkonw keyword: ' + args[0])

    def view(self, *args, **kargs):
        return self._view(*args, **kargs)

    def close(self, ):
        '''
        close window
        '''
        # simply close the window by calling Close of Frame
        import wx
        if self is not wx.GetApp().TopWindow:
            self.Close()
        else:
            self.close_figurebook()

    @share_doc_string_simple
    def cbar(self, name=None, *args,  **kargs):

        position = kargs.pop('position', (0.9, 0.1))
        size = kargs.pop('size', (0.05, 0.8))
        lsize = kargs.pop('lsize', 'default')
        lcolor = kargs.pop('lcolor', 'black')
        olsize = kargs.pop('olsize', 'default')
        olcolor = kargs.pop('olcolor', 'black')
        direction = kargs.pop('direction', 'v')

        update = kargs.pop('update', self._interactive_update)
        hold = kargs.pop('hold', True)
        f_page = self.get_page()
        if f_page is None:
            return
        ax = f_page.get_axes(self.isec())

        action = 'both'
        for k, p in enumerate(ax._caxis):
            if (not p.has_cbar() and (name is None or p.name == name) and
                    (action == 'on' or action == 'both')):
                p.show_cbar(ax,
                            offset=-0.1*k,
                            position=position,
                            size=size,
                            lsize=lsize,
                            lcolor=lcolor,
                            olsize=olsize,
                            olcolor=olcolor,
                            direction=direction[0])
                action = 'on'
            if (p.has_cbar() and (name is None or p.name == name) and
                    (action == 'off' or action == 'both')):
                action = 'off'
                p.hide_cbar()
        if update:
            ifigure.events.SendChangedEvent(
                self.book, w=self, useProcessEvent=True)
            ifigure.events.SendPVDrawRequest(
                self.book, wait_idle=True, refresh_hl=False)

    def _names(self, attr):
        f_page = self.get_page()
        if f_page is None:
            return
        try:
            ax = f_page.get_axes(self.isec())
            return [p.name for p in getattr(ax, attr)]
        except:
            return []

    @share_doc_string
    def xnames(self):
        return self._names('_xaxis')

    @share_doc_string
    def ynames(self):
        return self._names('_yaxis')

    @share_doc_string
    def znames(self):
        return self._names('_zaxis')

    @share_doc_string
    def cnames(self):
        return self._names('_caxis')

    @allow_interactive_call2
    def property(self, obj, *args, **kargs):
        from ifigure.widgets.artist_widgets import listparam, call_getter
        ret = obj.property_for_shell()
        tags = None
        if isinstance(ret, tuple):
            tab = ret[0]
            props = ret[1]
            if len(ret) == 3:
                tags = ret[2]
        else:
            tab = ['']
            props = [ret]
        if tags is None:
            tags = ['']*len(props)

        ll = []
        if 'list' in kargs or len(args) == 0:
            for tag, plist in zip(tags, props):
                for p in plist:
                    if tag == '':
                        ll.append(listparam[p][6])
                    else:
                        ll.append((tag, listparam[p][6]))
            return ll
#        if len(args) == 0: return

#        from matplotlib.cbook import is_string_like
        from ifigure.utils.cbook import isstringlike
        name = args[0]
        if not isstringlike(name):
            tag, name = name
        else:
            tag = ''
        args = args[1:]
        for tt, plist in zip(tags, props):
            for p in plist:
                if (listparam[p][6].upper() == name.upper() and
                        tt == tag):
                    ll.append(listparam[p])

        if len(ll) == 0:
            print('Property '+name + ' is not found')
            return
        if len(ll) > 1:
            print('More thant two properties are found by '+name)
            print('Please inform developer this problem')
            return
        # dprint1(ll)
        ll = ll[0]
        if len(args) == 0:
            ret = [call_getter(a, ll, tab=tag) for a in obj._artists]
            if len(ret) == 1:
                return ret[0]
            return ret
        else:
            actions = []
            vv = args[0]
            if len(obj._artists) == 1:
                vv = (vv,)
            from ifigure.utils.cbook import isiterable

            for a, v in zip(obj._artists, vv):
                name = ll[4]
                sw = ll[5]
                if sw > 9:
                    sw = sw-10
                if sw == 0:
                    action = UndoRedoArtistProperty(a, name, v)
                elif sw == 1:
                    action = UndoRedoFigobjProperty(a, name, v)
                elif sw == 8:
                    action = UndoRedoFigobjProperty(a, name, v, nodelete=True)
                elif sw == 2:
                    action = UndoRedoFigobjMethod(a, name, v)
                elif sw == 3:
                    action = UndoRedoFigobjMethod(a, name, v)
                    action.set_extrainfo(tag)
                actions.append(action)
            for rec in actions:
                rec.do()
            figobj = actions[-1].do2()
            figobj.set_bmp_update(False)
            self.draw()

    def size(self, *args):
        if len(args) == 0:
            return self.GetSize()
        try:
            value = (args[0], args[1])
        except:
            try:
                value = tuple(args[0])
            except:
                raise
        self.SetSize(value)
        return self

    @allow_interactive_call2
    def savefig(self, filename):
        import os
        filename = os.path.expanduser(filename)
        if self.num_page() == 0:
            return
        if filename[-4:] == '.eps':
            wc = 0
        elif filename[-4:] == '.pdf':
            if self.num_page() == 1:
                wc = 1
            else:
                wc = 7
        elif filename[-4:] == '.svg':
            wc = 2
        elif filename[-4:] == '.png':
            wc = 3
        elif filename[-5:] == '.jpeg':
            wc = 4
        elif filename[-4:] == '.gif':
            wc = 5
        else:
            assert False, "unknown file format"

        self.canvas.save_pic(ret=filename, wc=wc)

    @allow_interactive_call2
    def savedata(self, filename, **kargs):
        import os
        filename = os.path.expanduser(filename)
        if self.num_page() == 0:
            return
        if filename[-4:] != '.hdf':
            assert False, "filename must end with hdf"

        page = self.canvas._figure.figobj
        metadata = kargs.pop('metadata', OrderedDict())
        export_flag = kargs.pop('export_flag', {})
        verbose = kargs.pop('verbose', True)

        try:
            #  for standalone testing (when running python hdf_export_window.py)
            from ifigure.utils.hdf_data_export import (build_data,
                                                       hdf_data_export,
                                                       set_all_properties_all,
                                                       select_unique_properties_all)
        except:
            print("can not load hdf export tools")
            raise

        dataset, metadata, export_flag = build_data(page,
                                                    verbose=False,
                                                    metadata=metadata,
                                                    export_flag=export_flag)

        try:
            hdf_data_export(data=dataset,
                            metadata=metadata,
                            export_flag=export_flag,
                            filename=filename,
                            verbose=verbose)
            print('HDF export finished : '+filename)
        except:
            print("can not export  hdf dataset")
            raise



from __future__ import print_function
import time
import weakref
import threading
from six.moves import queue as Queue


lock = threading.Lock()
queues = []


def get_thread(return_queue=None):
    lock.acquire()
    new_queue = Queue.Queue()
    queues.append(new_queue)
    lock.release()
    return EventDrivenThread(queues[-1], return_queue), new_queue


def send_event(event):
    lock.acquire()
    for q in queues:
        q.put(event)
    lock.release()


def remove_queue(q):
    lock.acquire()
    if q in queues:
        queues.remove(q)
    lock.release()


class EventDrivenThread(threading.Thread):
    def __init__(self, queue, rqueue, *args, **kargs):
        super(EventDrivenThread, self).__init__(*args, **kargs)
        self._verbose = False
        self.queue = queue
        self.rqueue = rqueue
        self.events = {}
        self.start_event = self.name + 'start_action'
        self.finish_event = self.name + 'finish_action'
#        self.quit_event   = 'quit_thread'

    def bind(self, event, method):
        if event in self.events:
            self.events[event].append(method)
        else:
            self.events[event] = [method]

    def unbind(self, event, method):
        if event in self.events:
            if method in self.events[event]:
                self.events[event].remove(method)
            if len(self.events[event]) == 0:
                del(self.events[event])

    def bind_init(self, work, method):
        self.bind(work.get_full_path() + '.start', method)
        return work.get_full_path() + '.start'

    def run(self, *args, **kagrs):
        send_event(self.start_event)
        self.flag = True
        while (self.flag):
            if self.queue.empty():
                time.sleep(0.1)
            else:
                event = self.queue.get(False)
#               print event, self.name
#               print self.events
                if event in self.events:
                    methods = [m for m in self.events[event]]
                    for m in methods:
                        m(event)
#                       print 'unbind', event, m
                        self.unbind(event, m)
#                   print self.events
#                           print event, ' finished method  (', m.__name__,',', m.__self__.name, ') in ', self.name
#                       else:
#                           print event, ' attempted to perform method  (', m.__name__,',', m.__self__.name, ') in ', self.name

#                       print self.events
                time.sleep(0.1)
            if len(self.events) == 0:
                self.flag = False
        remove_queue(self.queue)
        send_event(self.finish_event)
        self.queue = None
        self.events = None
        if self.rqueue is not None:
            self.rqueue.put(self.name)
        self.rqueue = None
        print(('Exiting...', self.name))



from __future__ import print_function
from ifigure.utils.images2gif import writeGif
#from PIL import Image
from ifigure.utils.cbook import image_to_pil
import six
dither = 1
import tempfile
import os
import wx
import shutil

def save_animation(func, params, canvas, filename='animation.png', duration=0.2):

    path = tempfile.mkdtemp()
    images = []
    for k, p in enumerate(params):
        func(p)
        image = canvas.canvas.bitmap.ConvertToImage()
        fname = os.path.join(path, 'frame_'+str(k)+'.png')
        image.SaveFile(fname, wx.BITMAP_TYPE_PNG)
        images.append(fname)


    from apng import APNG
    APNG.from_files(images,  delay=int(duration)).save(filename)
    shutil.rmtree(path)



from __future__ import print_function
#
#   this is a skelton for a optimizer
#
#   It optimize parameter variables in model.param
#   to minimize a cost function
#
#   The cost function should be provilded by a user.
#
import wx
import weakref
import threading
from six.moves import queue as Queue
import ifigure.events


def modifier(worker, names, values):
    print('Running modifier')
    print(('working on model directory ...', worker))
    for name, value in zip(names, values):
        print(('appling new parameter...', name, value))
        worker.parameters.setvar(name, value)


solver = obj.get_parent()
tmodel = solver._model
rpath = tmodel.get_root_model().get_td_path(tmodel)
rname = tmodel.get_root_model()._name


def locate_cost_script():
    if solver._cost.startswith('.'):
        txt = solver.get_full_path()+solver._cost
    else:
        txt = solver._cost
    root = solver.get_root_parent()
    exec(root._name + ' = root')
    return eval(txt)


use_def_merger = solver._use_def_merger
renew_model_each_time = False


def run_workers(worker, sol):
    def submit(worker, queue, values):
        rmodel = worker.get_child(name=rname)
        if (renew_model_each_time and
                rmodel is not None):
            # print "Desotry", rmodel
            rmodel.destroy()
        rmodel = worker.get_child(name=rname)
        if rmodel is None:
            tmodel.get_root_model().duplicate(worker, rname, save_script_link=True)
            rmodel = worker.get_child(name=rname)
        model = rmodel.resolve_td_path(rpath)

        names = solver._pname
        modifier(worker, names, values)

        threads = model.Run(return_queue=queue)
        worker._thread_name = threads[0]

    def finish_job(method, idx, w, sol, queue):
        method(idx, w, sol)
        queue.put(w)
        solver._queue = None

    q = Queue.Queue()
    solver._queue = q
    cost_script = locate_cost_script()

    func_failed = False

    def func(values, w=worker, queue=q, ):
        aborted = False

        wx.CallAfter(submit, *(w, queue, values))
        try:
            while True:
                t = queue.get(True)
                if t == w._thread_name:
                    break
                elif t == 'abort job':
                    aborted = True
            if aborted:
                raise ValueError
                func_failed = True
            rmodel = w.get_child(name=rname)
            model = rmodel.resolve_td_path(rpath)
            c = cost_script.RunA(model)
            return c
        except Queue.Empty:
            solver._queue = None
            print('Failed to get response')
            raise ValueError

    # call scipy optimizer here
    from scipy.optimize import minimize

    try:
        kargs = solver._opt_kargs
        values = solver._init_value
        ret = minimize(func, values, **kargs)
    except:
        import traceback
        traceback.print_exc()
        solver._queue = None
        return
    if func_failed:
        return
    rmodel = w.get_child(name=rname)
    model = rmodel.resolve_td_path(rpath)

    if (model._finish_script == ')' or
            use_def_merger):
        model._finish_script = solver.merge_sol.get_full_path()
    # print 'calling merge_sol', model.do_finish
    wx.CallAfter(finish_job, model.do_finish, 0, w, sol, q)
    t = q.get(True)
    if t == 'abort job':
        print('abort recieved during postprocessing...')
    solver._queue = None
#    model.do_finish(0, w, sol)


sol_base = solver._sol
workers = solver.get_workers()
if len(workers) == 0:
    print('no workers')
    stop()
if solver._queue is not None:
    print('there is a waiting queue, exiting...')
    stop()
for w in workers:
    w._thread_name = ''

t = threading.Thread(target=run_workers, args=(workers[0], sol_base))
wx.CallAfter(t.start)



from __future__ import print_function
#################################################
#
#   NameList : PyData to hold a fortran namelist
#
#       Actual implementatin was done in
#       submodules
#################################################
import os
import re
import string
import wx
import shutil
from ifigure.utils.cbook import parseStr
from numpy import *
from collections import OrderedDict
from ifigure.mto.py_code import PyData
from ifigure.mto.py_contents import Namelist
######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.
#
#
#      Strong recommendation : make module "independent".
#
#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  moduling
#      order-dependency problem.
#
#   name of module
module_name = 'namelist'
class_name = 'namelist'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
menu = [("Import...", "onLoadFile", True),
        ("Export...", "onWriteFile", False),
        ("Update File", "onUpdateFile", False),
        ("Update Tree", "onUpdateTree", True),
        ("+TextEditor...", None,  False),
        ("View Original", "onOpenOrg", False),
        ("Edit File", "onOpenCurrent", False),
        ("!", None, False),
        ("+NameList...", None,  False),
        ("Add Section...", "onAddSec", False),
        ("Remove Section...", "onRmSec", False),
        ("Rename Section...", "onRenameSec", False),
        ("!", None, False)]

method = ['tree2txt', 'check_new_fileedit',
          'onLoadFile', 'onWriteFile', 'onOpenOrg',
          'onOpenCurrent', 'onUpdateFile', 'onUpdateTree',
          'onAddSec', 'onRmSec', 'onRenameSec',
          'init']
icon = 'text.png'
can_have_child = False
has_private_owndir = False
######################################################

modename = 'namelist_pathmode'
pathname = 'namelist_path'
extname = 'namelist_ext'


def init(self, *args, **kargs):
    obj = self.td
    obj.mk_owndir()
    nm = Namelist()
    obj.setvar0(nm)
    if 'src' not in kargs:
        self.onLoadFile()
    else:
        try:
            self.onUpdateTree()
        except:
            import traceback
            traceback.print_exc()


def split_arr(arr, pattern):
    ret = []
    for s in arr:
        ret = ret + re.split(pattern, s)

    ret2 = []
    for s in ret:
        if s != '':
            ret2 = ret2+[s]

    return ret2


def join_arr(arr, a, b):  # a='(', b=')'
    num = 0
    ret = []
    i = 0
    while i < len(arr):
        if arr[i] != a:
            ret = ret + [arr[i]]
        else:
            num = 0
            j = i+1
            tmp = arr[i]
            while (arr[j] == b and num == 0) is False:
                if arr[j] == a:
                    num = num+1
                if arr[j] == b:
                    num = num-1
                tmp = tmp+arr[j]
                j = j+1
            i = j
            tmp = tmp+arr[i]
            ret = ret + [tmp]
        i = i+1
    return ret


def join_arr2(arr, sep):
    ret = [arr[0]]
    i = 1
    while i < len(arr):
        flag = True
        for s in sep:
            if arr[i] == s:
                flag = False
        if flag:
            ret[-1] = ret[-1]+arr[i]
        else:
            ret = ret + [' ']
        i = i+1

    ret2 = []
    for s in ret:
        ret2 = ret2+[s.strip()]
    return ret2


def interpret_line(line):
    test = [line]
    test = split_arr(test, r'(")')
    test = split_arr(test, r'(,)')
    test = split_arr(test, r'(\()')
    test = split_arr(test, r'(\))')
    test = split_arr(test, r'(\s)')

    test = join_arr(test, '(', ')')
    test = join_arr(test, '"', '"')

    test = join_arr2(test, ', ')
    ret2 = []
    for s in test:
        if s != '':
            ret2 = ret2+[s]
    return ret2


def string_split_ig(s, sep):
    t = re.split(sep, s)
    id = range(len(t))
    id = reversed(id)
    for i in id:
        if t[i] == '':
            del t[i]
    return t


def eval_form2020(s):
    #form2020= '(5e16.9)'
    d = 16
    ret = []
    for i in range(5):
        if s[d*i:d*(i+1)] != '':
            ret.append(float(s[d*i:d*(i+1)]))
    return ret


def readandstrip(f):
    line = f.readline()
    line = line.rstrip("\r\n")
    return line


def load_array_form2020(f, size):
    k = 0
    arr = []
    while (k < size):
        arr = arr+eval_form2020(readandstrip(f))
        k = k+5
    return array(arr)


def load_matrix_form2020(f, size1, size2):
    k = 0
    arr = []
    while k < size1*size2:
        arr = arr+eval_form2020(readandstrip(f))
        k = k+5
    return array(arr).reshape(size1, size2)


def load_file(file=None):
    #   file='/home/shiraiwa/PycharmProjects/ifigure/sample/genray.in'
    def get_line(f):
        line0 = f.readline()
#      print line0
        if not line0:
            return False, '', ''
        if line0.find('!') != -1:
            line0 = line0[:line0.find('!')]
#         print line0
        line0 = line0.rstrip("\r\n")
        line0 = ' '.join(line0.split())
        if line0.endswith('/'):
            line = line0[:-1]
        else:
            line = line0
        return True, line, line0
    f = open(file, 'r')

    # namelist section
    nm = Namelist()
    sec = None
    end_flag = False
    while 1:

        flag, line, line0 = get_line(f)
#      print line0
        if not flag:
            break
        if line == '':
            continue

        if (line.startswith('&') and
                line[:4].lower() != '&end'):
            while line[1] == ' ':
                line = line[:1] + line[2:]
            s = string_split_ig(line, ' |,')
            sec = s[0][1:]
            #print("entring section", sec)
            nm[sec] = OrderedDict()
            if len(s) > 1:
                line = ' '.join(s[1:])
            else:
                continue
        if sec is None:
            continue  # skip unitl first &*** starts

        sall = interpret_line(line)
        i = 0
        while i < len(sall):
            if sall[i].find('=') != -1:
                if sall[i] == '=':    # '='
                    sall[i-1] = sall[i-1]+'='
                    del sall[i]
                    continue
                if sall[i].startswith('='):  # '=value'
                    sall[i-1] = sall[i-1]+'='
                    sall[i] = sall[i][1:]

                    continue
                if sall[i].endswith('='):
                    print(sall[i])
                    i = i+1
                    continue  # 'name='
                k = sall[i].split('=')
                sall[i] = k[0]+'='
                sall.insert(i+1, k[1])
            i = i+1

        for s in sall:
            if s[:4].lower() == '&end':
                sec = None
#           print('breaking')
                break
            if s.find('=') != -1:
                k = s.split('=')
                varname = k[0]
#          s[0]=k[1]
#          print s
#          if s[0] is '': del s[0]
#          print 'create dict key', sec, varname
                nm[sec][varname] = []
                if s.endswith('/'):
                    sec = None
                continue
        # for lines without 'xxxx = '

            nm[sec][varname] = nm[sec][varname]+parseStr(s)
        if line0.endswith('/'):
            sec = None
    f.close()
    return nm


def load_namelistfile(obj):
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)

    if file == '':
        return

    for name, child in obj.get_children():
        child.destroy()

    nm = load_file(file)
    print(('reading file', file))
    obj.setvar0(nm)
    mtime = os.path.getmtime(file)
    obj.setvar('namelist_mtime', mtime)


def tree2txt(self):
    '''
    make tree to txt and update owndir file
    '''

    td = self.td
    var0 = td[:]
    txt = []
    for key0 in var0:

        txt.append(' &'+key0+'\n')
        var = var0[key0]

        for key in var:
            line = ' '+key+' = '

            for el in var[key]:
                if len(line) + len(str(el)) > 73:
                    txt.append(line+'\n')
                    line = '    '
                line = line + ' '+str(el)
            txt.append(line+'\n')
        txt.append(' &end\n \n')

    txt2 = ''.join(txt)

    return txt2


def check_new_fileedit(self, no_dialog=False, force_load=False):
    '''
    check if file is edited since the last loading

    by default, it asks if a user wants to reload
    otherwise, it returns status. In the latter case,
    a program needs to reload by itself later 
    '''
    obj = self.td
    if (obj.get_mtime(modename='namelist_pathmode', pathname='namelist_path') >
            obj.getvar('namelist_mtime')):
        if not no_dialog:
            dlg = wx.MessageDialog(None,
                                   'File is newer than tree data, Do you want to reload?',
                                   'Old tree data',
                                   wx.OK | wx.CANCEL)
            if dlg.ShowModal() == wx.ID_OK:
                self.onUpdateTree()
                dlg.Destroy()
                return True
            else:
                return False
        else:
            if force_load:
                #               print('loading namelist from file')
                self.onUpdateTree()
                return True
            else:
                return False
    return True


def onUpdateTree(self, e=None):
    obj = self.td
    load_namelistfile(self.td)


def onLoadFile(self, e=None, file=''):
    from ifigure.utils.addon_utils import onLoadFile

    if file != '':
        self.td.set_path_pathmode(file, modename, pathname, extname)
        ret = True
    else:
        ret = onLoadFile(self.td, message="Select Namelist File",
                         modename=modename,
                         pathname=pathname,
                         extname=extname)
    if ret:
        load_namelistfile(self.td)


def onWriteFile(self, e=None, filename=''):
    from ifigure.utils.addon_utils import onWriteFile

    txt = self.tree2txt()
    onWriteFile(self.td, filename=filename,
                txt=txt,
                message='Enter Namelist File Name')


def onUpdateFile(self, e=None, *args, **kargs):
    obj = self.td
    file = self.td.path2fullpath(modename=modename,
                                 pathname=pathname)
    file_check = os.path.join(obj.owndir(), obj.name)

    if file == '':
        file = file_check
        try:
            self.onWriteFile(filename=file)
            obj.set_path_pathmode(file, modename, pathname, extname)
        except:
            print('file save failed')
            traceback.print_exc()
    else:
        self.onWriteFile(filename=file)


def onOpenOrg(self, e=None, *args, **kargs):
    from ifigure.utils.addon_utils import onOpenOrg
    onOpenOrg(self.td)


def onOpenCurrent(self, e=None, *args, **kargs):
    from ifigure.utils.addon_utils import onOpenCurrent
    onOpenCurrent(self.td, modename=modename,
                  pathname=pathname)


def onAddSec(self, e):
    import ifigure.widgets.dialog as dialog

    parent = e.GetEventObject()
    ret, name = dialog.textentry(parent,
                                 "Enter section name",
                                 "Add Section...", 'section')
    if not ret:
        return

    var0 = self.td.getvar0()
    if not name in var0:
        var0[name] = Namelist()
        app = wx.GetApp().TopWindow
        app.proj_tree_viewer.update_content_widget(self.td)


def onRmSec(self, e):
    import ifigure.widgets.dialog as dialog

    parent = e.GetEventObject()
    var0 = self.td.getvar0()
    ret, name = dialog.textselect(parent,
                                  "Select section name",
                                  "Remove Section...",
                                  choices=list(var0.keys()))
    if not ret:
        return

    del var0[name]
    app = wx.GetApp().TopWindow
    app.proj_tree_viewer.update_content_widget(self.td)


def onRenameSec(self, e):
    import ifigure.widgets.dialog as dialog
    parent = e.GetEventObject()
    var0 = self.td.getvar0()
    ret, name = dialog.textselect(parent,
                                  "Select section name",
                                  "Rename Section...",
                                  choices=list(var0.keys()))
    if not ret:
        return
    ret, name1 = dialog.textentry(parent,
                                  "Enter new section name",
                                  "Rename Section...", name+'_renamed')
    if not ret:
        return

    var0[name1] = var0[name]
    del var0[name]
    app = wx.GetApp().TopWindow
    app.proj_tree_viewer.update_content_widget(self.td)



from __future__ import print_function

'''
   Setting Parser

   multipurpse setting file parser, which support three types of
   text input file.
   As for the field value, tt tries to evaluate string.
   This convert '16' to a number. If it is necessary to have 
   string '"16"' should be used. If it can not convert, the
   field value is considered as string
       0)  before any field start, global field can be added.
           it can be used for "version" field for example.
       1)  random field file
               x=3
               y=4
       2)  list of data witten in the same format
               student
                 name = 'bob'
                 age = 10
               end
               student
                 name = 'jay'
                 age = 14
               end
       2)  fortran namelist type file
               network
                 remote_host = 'cmodws60'
                 port = 22
               end
               interface
                  default_width = 300
                  default_height = 400
               end
       3) field name will be all converted to lower case.


   useage:
       filename = 'some file name'
       sp = SettingParser(rule_mode=1)
       sp.set_rule('student', {'name':'', 'age':16})
       print sp.read_file(filename)
'''

import os
from ifigure.utils.cbook import isiterable


class SettingParser(object):
    def __init__(self, rule_mode=1):
        '''
        rule_mode = 1: read as many set of parameters based on rule
        rule_mode = 2: there should be only one set for one rule
                       newer data will overwrite the exising data
        '''
        object.__init__(self)
        self.rule = {}
        self.rule_mode = rule_mode
        self.rule_nocheck = {}
        self.rule_usedefault = {}

    def set_rule(self, name, rule, nocheck=False, usedefault=True):
        self.rule[name] = rule
        self.rule_nocheck[name] = nocheck
        self.rule_usedefault[name] = usedefault
        return self

    def omit_comment(self, line):
        if line.find('#') != -1:
            line = line[0:line.find('#')]
        if line.find('!') != -1:
            line = line[0:line.find('#')]
        return line

    def split_line(self, _line):
        _arr = _line.split('=')
        _name = _arr[0].strip()
        _value = '='.join(_arr[1:]).strip()
        try:
            value = eval(_value)
        except:
            value = _value
        return _name.lower(), value

    def read_set(self, lc, lines, name, nocheck=False, usedefault=True):
        if usedefault:
            rule = dict(self.rule[name])
        else:
            rule = {}
        while lc < len(lines):
            line = lines[lc]
            line = self.omit_comment(line)
            if line.upper().find('END') != -1:
                lc = lc+1
                break
            if line.find('=') == -1:
                lc = lc+1
                continue

            name2, value = self.split_line(line)
            if nocheck or (name2 in rule):
                rule[name2] = value
            else:
                print(('Undefined field :', name2, ' in rule ' + name))
            lc = lc+1
        return lc, rule

    def read_file(self, file='', data=None, lines=None):
        '''
        if lines is given, it analyze line, otherwise,
        it parses a file 
        '''
        if lines is None:
            try:
                fid = open(file, 'r')
                lines = fid.readlines()
                fid.close()
            except IOError:
                return False, {}
        lines = [line.strip('\n\r') for line in lines]
#        print lines
        if data is None:
            data = {}
            if self.rule_mode == 1:
                for key in self.rule:
                    data[key] = []
        lc = 0
        while lc < len(lines):  # lc = line counter
            line = lines[lc]
            line = self.omit_comment(line)
            if line.find('=') == -1:
                if line.strip() in self.rule:
                    rname = line.strip()
                    lc = lc+1
                    lc, value = self.read_set(lc, lines, rname,
                                              nocheck=self.rule_nocheck[rname],
                                              usedefault=self.rule_usedefault[rname])
                    if rname in data:
                        data[rname].append(value)
                    else:
                        data[rname] = value
                    continue
                else:
                    lc = lc+1
                    continue
            name, value = self.split_line(line)
            data[name] = value
            lc = lc+1
        return True, data

    def write_subsection(self, lines, name, data):
        lines.append(name)
        for key in data:
            if isinstance(data[key], str):
                lines.append(key + ' = "' + str(data[key])+'"')
            else:
                lines.append(key + ' = ' + str(data[key]))
        lines.append('end')
        return lines

    def write_file(self, data, file=None):
        lines = []
        for key in data:
            if isinstance(data[key], dict):
                lines = self.write_subsection(lines, key, data[key])
            elif isinstance(data[key], list):
                for item in data[key]:
                    lines = self.write_subsection(lines, key, item)
            elif isinstance(data[key], tuple):
                for item in data[key]:
                    lines = self.write_subsection(lines, key, item)
            elif isinstance(data[key], str):
                lines.append(key + ' = "' + str(data[key])+'"')
            else:
                lines.append(key + ' = ' + str(data[key]))

        if file is None:
            for l in lines:
                print(l)
            return
        fid = open(file, 'w')
        for l in lines:
            fid.write(l+'\n')
        fid.close()


class iFigureSettingParser(SettingParser):
    def mname2file(self, mname):
        from ifigure.ifigure_config import rcdir, ifiguredir, resourcedir
#        print mname, resourcedir
#        def_file = os.path.join(*([ifiguredir] + mname.split('.')))
        def_file = os.path.join(*([resourcedir] + mname.split('.')))
        user_file = os.path.join(*([rcdir] + mname.split('.')))
        return def_file, user_file

    def mk_filepath_dir(self, filename):
        t = filename
        if os.path.exists(t):
            return
        if not os.path.exists(os.path.dirname(t)):
            self.mk_filepath_dir(os.path.dirname(t))
        else:
            print(('making directory', t))
            os.mkdir(t)

    def read_setting(self, mname, fromDefault=False):
        def_file, user_file = self.mname2file(mname)
#        print def_file
#        print user_file

        flag, def_v = self.read_file(def_file)
        if (os.path.exists(user_file) and
                not fromDefault):
            flag, user_v = self.read_file(user_file)
            try:
                #                print user_v['version'] , def_v['version']
                if user_v['version'] < def_v['version']:
                    self.write_setting(mname, def_v)
                    return def_v
                else:
                    return user_v
            except:
                self.write_setting(mname, def_v)
                return def_v
        else:
            self.write_setting(mname, def_v)
            return def_v

    def write_setting(self, mname, data):
        user_file, user_file = self.mname2file(mname)
        self.mk_filepath_dir(os.path.dirname(user_file))
        self.write_file(data, user_file)


if __name__ == '__main__':

    lines = ['version = 1',
             'port = 22',
             'host = transport']
    sp = SettingParser()
    print(sp.read_file(lines=lines))

    lines = ['version = 1',
             'student',
             'name = "bob"',
             'age = 17',
             'end',
             'student',
             'name = "jones"',
             'age = 14',
             'end',
             'student',
             'name = "anne"',
             'age = 15',
             'end']

    sp = SettingParser(rule_mode=1)
    sp.set_rule('student', {'name': '', 'age': 16})
    print(sp.read_file(lines=lines))

    lines = ['version = 1',
             'type = "proxy"',
             'connection',
             "server = 'cmodws60.psfc.edu'",
             "port = 10002",
             "end",
             "connection",
             "server = 'cmodws59.psfc.mit.edu'",
             "port = 10002",
             "end"]

    sp = SettingParser(rule_mode=1)
    sp.set_rule('connection', {}, nocheck=True)
    flag, data = sp.read_file(lines=lines)
    print(data)
    sp.write_file(data)

    lines = ['house',
             'rooms = 5',
             'garage = 2',
             'end',
             'yard',
             'type  = "lawn"',
             'end',
             'toy',
             'name = "train"',
             'end']

    sp = SettingParser(rule_mode=2)
    sp.set_rule('house', {})
    sp.set_rule('yard', {})
    sp.set_rule('toy', {})
    print(sp.read_file(lines=lines))

    lines = ['version = 1',
             'house',
             'rooms = 5',
             'garage = 2',
             'end',
             'yard',
             'type  = "lawn"',
             'end',
             'toy',
             'NAME = "1,1,1,1"',
             'end']

    sp = SettingParser(rule_mode=2)
    sp.set_rule('house', {}, nocheck=True)
    sp.set_rule('yard', {}, nocheck=True)
    sp.set_rule('toy', {}, nocheck=True)
    flag, data = sp.read_file(lines=lines)
    print(data)
    sp.write_file(data)



from __future__ import print_function
#  Name   :fig_mds
#
#          container to create and hold figures
#          for mdsplus system
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#
# *******************************************
#     Copyright(c) 2012- S. Shiraiwa
# *******************************************
import traceback
from collections import OrderedDict
import numpy as np

from ifigure.widgets.canvas.file_structure import *

import logging
import os
import weakref
import time
import threading
import ifigure.utils.cbook as cbook
import ifigure.utils.pickle_wrapper as pickle

from ifigure.mto.treedict import TreeDict
from ifigure.mto.fig_obj import FigObj
from ifigure.mto.py_script import PyScript
from ifigure.mto.py_code import PyCode
from ifigure.mto.fig_grp import FigGrp


import ifigure.utils.geom as geom
import ifigure.events

from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle

from ifigure.mdsplus.filed_ndarray import FiledNDArray, FiledNDArrayAutoDel
from ifigure.mdsplus.mds_job import MDSjob
import wx
#
#  debug setting
#
import ifigure.utils.debug as debug
from .utils import parse_server_string
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('FigMds')
script_file_name = 'mdsscript.py'

panel_text = {'timetrace': 'required y\noptional x \n(x should monotonically increase)',
              'stepplot': 'required y\noptional x \n(x should monotonically increase)',
              'plot': 'required y\noptional x, xerr, yerr',
              'contour': 'required z\noptional x, y',
              'image': 'required z\noptional x, y',
              'axline': 'required either x or y',
              'axspan': 'required either x or y',
              'text': 'required s\noptional x, y',
              'surface': 'required z\noptional x, y',
              'noplot': ''}

required_variables = {'plot': ('y', 'x'),
                      'timetrace': ('y', 'x'),
                      'stepplot': ('y', 'x'),
                      'contour': ('z', 'x', 'y'),
                      'image': ('z', 'x', 'y'),
                      'surface': ('z', 'x', 'y'),
                      'axline': ('x', 'y'),
                      'axspan': ('x', 'y'),
                      'text': ('s', 'x', 'y'),
                      'noplot': tuple()}


def read_scriptfile(file):
    f = open(file, 'r')
    txt = ''
    while 1:
        line = f.readline()
        if not line:
            break
        txt = txt+line
    f.close()
    return txt


def write_scriptfile(file, txt):
    f = open(file, 'w')
    f.write(txt)
    f.close()
    return txt


def build_init_local_global_session(book):
    ana = MDSsession()
    server = book.getvar('mdsplus_server')

    if server is not None:
        s, p, t = parse_server_string(server)
#         job0 = MDSjob('reset', server)
#         ana.add_job([job0], 'reset')
        ana.add_job(
            [MDSjob('init', 'reset_private();reset_public();1')], 'reset_lg')
        ana.isec = 0
        ana.ishot = 0
        ana.ipage = 0
        ana.shot = 0
        ana.postprocess_done = True
        return ana
    else:
        return None


def add_connect_session(book, ana_group):
    ana = ana_group[0]
    expr = book.getvar('global_tdi')
    if expr is not None and expr != '' and book._use_global_tdi:
        job1 = MDSjob('value', expr, 'global_tdi')
#        print 'adding global tdi', expr
        ana_group.global_tdi = job1
    else:
        ana_group.global_tdi = None
    server = book.get_figbook().getvar('mdsplus_server')
    job0 = MDSjob('connection_mode', server)
    ana.add_job([job0], 'connection_mode', idx=0)
#    ana.postprocess_done = True
    return ana_group


def build_current_shot_session(book):
    ana = MDSsession()
    server = book.getvar('mdsplus_server')

    if server is not None:
        s, p, t = parse_server_string(server)
#         print s, p, t
        if t == '':
            # if tree name for shot number is not
            # specified
            return None

#         job0 = MDSjob('connection_mode', server)
#         ana.add_job([job0], 'connection_mode')
#        print t

#  this seems so slow
#         job1 = MDSjob('open', t,  0)
#         ana.add_job([job1], 'connection')
#         ana.add_job([MDSjob('value', '$SHOT')], 'shot')

#  equivalent to MDSVALUE('current_shot("treename")')
        ana.add_job([MDSjob('value', 'current_shot("'+t + '")')], 'shot')
        ana.isec = -1
        ana.ishot = -1
        ana.ipage = -1
        ana.shot = 0
        return ana
    else:
        return None


class MDSJobSet(list):
    def __init__(self, *args, **kargs):
        list.__init__(self, *args, **kargs)
        self.can_compress = True


class MDSsession(object):
    def __init__(self):
        self.jobs = MDSJobSet()
        self.jobnames = []
        self.result = {}
        self.ipage = -1
        self.w_id = -1
        self.pool = None
        self.done = False
        self.finished_count = -1
        self.job_count = 0
        self.running = False
        self.status = False
        self.postprocess = None
        self.skipped = False
        self.postprocess_done = False

#       self.batchmode = True ## batch mode submit all job at once
    def __repr__(self):
        return 'MDSSession: '+str([job for job in self.jobs])

    def add_job(self, jobs, name, idx=None):
        if idx is None:
            self.jobs.append(jobs)
            self.jobnames.append(name)
        else:
            self.jobs.insert(idx, jobs)
            self.jobnames.insert(idx, name)
        self.result[name] = None
        self.job_count = self.job_count + 1

    def get_result(self, name):
        i = self.jobnames.index(name)
        return self.result[i]

    def set_result(self, flag, value):
        self.result = value
        extra = {}
        for key in self.result:
            if isinstance(self.result[key], FiledNDArray):
                self.result[key].__class__ = FiledNDArrayAutoDel
                try:
                    extra[key+'_catalog'] = self.result[key]
                    self.result[key] = self.result[key].restore()
                except:
                    print(traceback.format_exc())
                    self.result['error message'] = ['failed to file transfar']
        for key in extra:
            self.result[key] = extra[key]
        self.done = True
        self.running = False
        self.status = flag
        if len(value['error message']) != 0:
            self.status = False

    def print_jobs(self):
        for job in self.jobs:
            for j in job:
                j.print_job()

    def txt_jobs(self):
        v = []
        for job in self.jobs:
            for j in job:
                v.append(j.txt_job())
        return v


        # this is an option to generate FigXXX
def_plot_options = {'plot': (('',), {'color': 'auto',
                                     'markersize': 3,
                                     'markerfacecolor': 'auto',
                                     'markeredgecolor': 'auto'}),
                    'timetrace': (('',), {'color': 'auto',
                                          'markersize': 3,
                                          'markerfacecolor': 'auto',
                                          'markeredgecolor': 'auto'}),
                    'stepplot': (('',), {'color': 'auto',
                                         'markersize': 3,
                                         'markerfacecolor': 'auto',
                                         'markeredgecolor': 'auto'}),
                    'contour': ((7,), dict()),
                    'image': (tuple(), dict()),
                    'axline': (tuple(), dict()),
                    'axspan': (tuple(), dict()),
                    'text': (tuple(), dict()),
                    'surface': (tuple(), dict()), }


class FigMds(FigGrp):
    def __init__(self, *args, **kargs):
        if 'dwplot' in kargs:
            dwplot = kargs['dwplot']
            del(kargs['dwplot'])
        else:
            dwplot = {}
        self._mds_debug = False
        self._var_mask = []
        self._plot_type = 'noplot'
        self._plot_type = 'timetrace'
        self._session_editor = None
        self._shot = {}  # place to store analized shot
        self._shots_expr = []
        self._shot_expr = []  # this may not necessary?
        self._analysis_flag = dict()  # stores if analysis for ishot succeeded
        # a place holder most likely overwritten by
        self._color_order = ['black', 'red']
        # scope
        self._debug_ana = False
        super(FigMds, self).__init__(*args, **kargs)
        self.setvar('plot_options', def_plot_options.copy())

        # call import even when dwplot is empty so that
        # it set up variables in _var
        self.import_dwplot(dwplot)

    def isDraggable(self):
        return False

    def import_dwplot(self, dwplot):
        def read_dwplot(name, dwplot):
            if name in dwplot:
                return dwplot[name]
            return ''

        self._plot_type = 'timetrace'
        self.setvar("experiment", read_dwplot('experiment', dwplot))
        self.setvar("default_node",  read_dwplot('default_node', dwplot))

        # this is an option to generate FigXXX
        plot_options = self.getvar('plot_options')
        if 'show_mode' in dwplot:
            if int(dwplot['show_mode']) == 2:
                plot_options['timetrace'] = (('-o',),
                                             def_plot_options['timetrace'][1].copy())
                plot_options['stepplot'] = (('-o',),
                                            def_plot_options['stepplot'][1].copy())

            elif int(dwplot['show_mode']) == 1:
                plot_options['timetrace'] = (('s',),
                                             def_plot_options['timetrace'][1].copy())
                plot_options['stepplot'] = (('s',),
                                            def_plot_options['stepplot'][1].copy())
        #plot_options['timetrace'] = plot_options['plot']
        self.setvar('plot_options', plot_options)

        data = OrderedDict()
        varnames = required_variables[self._plot_type]
        for name in varnames:
            data[name] = read_dwplot(name, dwplot)
        self.setvar("mdsvars", data)
        self.setvar("title", read_dwplot('title', dwplot))
        self.setvar("event", read_dwplot('event', dwplot))

        stepplot = read_dwplot('step_plot', dwplot)
        if stepplot != '' and int(stepplot) == 1:
            self._plot_type = 'stepplot'

        self.setp('dw_extra', read_dwplot('global_defaults', dwplot))

        range_data = ((read_dwplot('xmin', dwplot),
                       read_dwplot('xmax', dwplot)),
                      (read_dwplot('ymin', dwplot),
                       read_dwplot('ymax', dwplot)))
        self._default_xyrange = {'flag': (False, False),
                                 'tdi': range_data,
                                 'value': ((-1, 1), (-1, 1))}
        self.apply_mdsrange()

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'mdsplus.png')
        return [idx]

    @classmethod
    def isFigMds(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'mdsplus'

    @classmethod
    def property_in_palette(self):
        #        return ["mdsfiguretype","mdssource","mdsevent", "mdsrange"]
        return ["mdssource", ]

    def get_mdsevent(self, a=None):
        return self.getvar('event')

    def set_mdsevent(self, value, a=None):
        self.setvar('event', value)
        # update viewers event wait list (if it is opened by mdsscope)
        if value != '':
            book = self.get_figbook()
            if book is None:
                return
            import wx
            v = wx.GetApp().TopWindow.find_bookviewer(book)
            if v is None:
                return
            if hasattr(v, 'make_event_dict'):
                v.make_event_dict()

    def get_artist_extent(self, a=None):
        ax = self.get_container()
        if ax is None:
            return [None]*4
        x1, y1 = ax.transAxes.transform((0, 0))
        x2, y2 = ax.transAxes.transform((1, 1))
        return [x1, x2, y1, y2]

    def canvas_menu(self):
        return []  # ("Data Setting...",  self.onDataSetting, None)]

    def tree_viewer_menu(self):
        return ([("Data Setting...",  self.onDataSetting, None),
                 ("Script...",  self.onEditScript, None),
                 ("---",  None, None)] +
                super(FigGrp, self).tree_viewer_menu())

    def onDataSetting(self, evt, noapply=False):
        if self._session_editor is not None:
            try:
                self._session_editor.Raise()
                return
            except:
                pass
        from ifigure.mdsplus.dlg_mdssession import DlgMdsSession
        p = evt.GetEventObject().GetTopLevelParent()
        data = self.getvar('mdsvars')
        cb = self.data_setting_closed
        self._session_editor = DlgMdsSession(p, data=data,
                                             figmds=self,
                                             cb=cb,
                                             noapply=noapply)
        self._session_editor.set_required_variables(
            required_variables[self._plot_type])
        evt.Skip()

    def data_setting_closed(self, value):
        self.set_mdsfiguretype(value[0], None)
        self.set_mdsrange(value[1], None)
        self.set_mdsevent(value[2], None)

    def applyDlgData(self, data):
        self.setvar('mdsvars', data)

    def onEditScript(self, evt):
        if not self.has_owndir():
            self.mk_owndir()
        filename = 'mdsscript.py'
        self.setvar('pathmode', 'owndir')
        self.setvar('path', filename)
        filepath = os.path.join(self.owndir(), filename)
        if not os.path.exists(filepath):
            fid = open(filepath, 'w')
            fid.close()
        super(FigMds, self).onEditScript(evt)

    def clean_mdsdata(self):
        for name, child in self.get_children():
            child.destroy()

    def mdssession_job(self, shot=None, dwglobal=None,
                       startup=None):
        '''
        evaluate mds expression 
        shot : shot list
        dw_global : figobjparam which stores global setting
        '''
        if shot == 'm':
            shot = -1
        if shot == 'n':
            shot = -1

        def get_mdsexpr(name, ls, gs):
            if name in ls and not name in self._var_mask:
                exp = ls[name]
            else:
                exp = ''
            if (exp is None or
                    exp == '') and (gs is not None):
                exp = gs[name]
            if exp is None:
                return ''
            if sum([len(x.strip()) for x in exp.split('\n')]) == 0:
                exp = ''
            return exp

        def check_node_in_exp(exp, node):
            node = node.strip()
            if len(node) == 0:
                return True
            if node[0] != '\\':
                return True
            if node.find('::') == -1:
                return True
            try:
                if node[1:node.index('::')].upper() == exp.upper():
                    return True
                return False
            except:
                return False

        exp = get_mdsexpr('experiment', self.getvar(), dwglobal)
        node = ''
        if (self.getvar('default_node') is None or
                self.getvar('default_node') == ''):
            if (self.getvar('experiment') == '' or
                    self.getvar('experiment') is None):
                # if both default node and experiment is blank
                node = get_mdsexpr('default_node', self.getvar(), dwglobal)
            else:
                node = ''
        else:
            node = self.getvar('default_node')
        check = exp.upper()
#       for item in dwglobal['_flag']:
#            print 'need to run this', item

        # special handleing for x, y, and z

        y_expr = get_mdsexpr('y', self.getvar('mdsvars'), dwglobal)
        x_expr = get_mdsexpr('x', self.getvar('mdsvars'), dwglobal)
        z_expr = get_mdsexpr('z', self.getvar('mdsvars'), dwglobal)
        title_expr = get_mdsexpr('title', self.getvar(), dwglobal)

        dim_of_expr = []
        if (x_expr == '' or x_expr == None) and y_expr != '':
            #            tmp = y_expr.split('\n')[-1]
            #            if tmp.endswith(';'): tmp = tmp[:-1]
            #            if tmp.find('=') != -1:
            #               tmp = tmp[:(tmp.find('='))]
            #            x_expr = 'dim_of('+tmp+')'
            #            dim_of_expr.append('x')
            #            x_expr = 'y,0'+y_expr
            x_expr = 'dim_of(_piscopevar)'
            dim_of_expr.append('y')
#            y_expr = '_piscopevar=('+y_expr+')'
        elif ((x_expr == '' or x_expr == None) and
              (y_expr == '' or y_expr == None) and
              (z_expr != '')):
            #            dim_of_expr.append('x')
            #            dim_of_expr.append('y')
            #            x_expr = 'z,0'+z_expr
            #            y_expr = 'z,1'+z_expr

            #            tmp = z_expr.split('\n')[-1]
            #            if tmp.endswith(';'): tmp = tmp[:-1]
            #            if tmp.find('=') != -1:
            #               tmp = tmp[:(tmp.find('='))]
            #            x_expr = 'dim_of('+tmp+',0)'
            #            y_expr = 'dim_of('+tmp+',1)'

            x_expr = 'dim_of(_piscopevar)'
            y_expr = 'dim_of(_piscopevar,1)'
            dim_of_expr.append('z')
#            z_expr = '_piscopevar=('+y_expr+')'

        def_names = ['x', 'y', 'z']

        vars = self.getvar('mdsvars')
        other_expr = [(key, vars[key]) for key in vars if not key in def_names]
        pos_txt = self.getvar('posvars')

        ana = MDSsession()
        if not (x_expr == '' and y_expr == '' and z_expr == '' and
                len(other_expr) == 0):
            server = self.get_figbook().getvar('mdsplus_server')
            job1 = [MDSjob('open', exp,  shot)]
            ana.add_job(job1, 'connection')
#            if node == '': node = "\\"+exp.split(',')[-1]+"::TOP"
#            print '**** default node *****', node
            if node != '':
                job2 = MDSjob('defnode', node)
                ana.add_job([job2], 'def_node')
#            if self.getvar('posvars') is not None:
#                 ana.add_job([MDSjob('value', self.getvar('posvars'))], 'pos')
            if pos_txt is None:
                ana.add_job(
                    [MDSjob('value', '_shot = '+str(shot))],    'shot_number')
            else:
                ana.add_job([MDSjob('value', pos_txt + ';' + '_shot = '+str(shot))],
                            'shot_number')
#            print ana.jobs, ana.jobnames
            if dwglobal is not None:
                for item in dwglobal['_flag']:
                    ana.add_job([MDSjob('value', dwglobal[item])], item)
#                print MDSjob('value', dwglobal[item], item)
            if z_expr != '':
                ana.add_job([MDSjob('value', z_expr)], 'z')
            if y_expr != '':
                if 'y' in dim_of_expr:
                    #                    ana.add_job([MDSjob('dim_of', y_expr)], 'y')
                    ana.add_job([MDSjob('valuesig', y_expr)], 'y')
                else:
                    ana.add_job([MDSjob('value', y_expr)], 'y')
            if x_expr != '':
                if 'x' in dim_of_expr:
                    ana.add_job([MDSjob('valuesig', x_expr)], 'x')
                else:
                    ana.add_job([MDSjob('value', x_expr)], 'x')
            for key, expr in other_expr:
                if expr.strip() != '':
                    ana.add_job([MDSjob('valuesig', expr)], key)
        if self._debug_ana:
            print(ana)
        can_compress = True if self._plot_type == 'timetrace' else False
        txt = self.make_script(shot, startup=startup)
        if txt is not None:
            if not self.get_script_local():
                ana.add_job([MDSjob('script_txt', txt)], 'ans')
            else:
                can_compress = False

        # if title_expr is not enclosed by '"', then put it.
        if title_expr == '':
            title_expr = '""'
            jobname = 'novalue'
        else:
            if title_expr[0] != '"' and title_expr[-1] != '"':
                title_expr = '"'+title_expr + '"'
            jobname = 'value'
        if True:
            shots_expr = self.make_shots_expr()
            if pos_txt is None:
                ana.add_job(
                    [MDSjob(jobname, '_shots = "'+shots_expr+'"')], 'shots')
            else:
                ana.add_job([MDSjob(jobname,
                                    pos_txt + ';' + '_shots = "'+shots_expr+'"')],  'shots')
            ana.add_job([MDSjob(jobname, title_expr)], 'title')
        ana.jobs.can_compress = can_compress
        return ana

    def make_script(self, shot, short=False, startup=None):
        if (self.has_owndir() and
                self.hasvar('path')):
            fname = os.path.join(self.owndir(), self.getvar('path'))
            if os.path.exists(fname):
                txt = read_scriptfile(fname)
                if short:
                    return txt
                txt = 'shot = ' + str(shot) + '\n'+txt
                if self.getvar('posvars') is not None:
                    txt = self.getvar('posvars')+'\n'+txt
                if startup is not None:
                    txt = read_scriptfile(startup) + '\n' + txt
                return txt

    def erase_mdsdata(self):
        for name, child in self.get_children():
            child.destroy()
        self.set_bmp_update(False)

    def reset_shots_expr(self):
        self._shots_expr = []

    def append_shots_expr(self, ss):
        self._shots_expr.append(str(ss))

    def make_shots_expr(self):
        return ','.join(self._shots_expr)

    def postprocess_data(self, ana, viewer,
                         color_order=['red', 'blue', 'yellow'],
                         ):
        def set_auto_color_to_color(kargs, name, col):
            if (not name in kargs) or (kargs[name] == 'auto'):
                kargs[name] = col
            return kargs

        from ifigure.mto.fig_plot import FigPlot, TimeTrace, StepPlot
        from ifigure.mto.fig_contour import FigContour
        from ifigure.mto.fig_image import FigImage
        from ifigure.mto.fig_surface import FigSurface
        from ifigure.mto.fig_axline import FigAxline
        from ifigure.mto.fig_axspan import FigAxspan
        from ifigure.mto.fig_text import FigText

        self._color_order = color_order
        dprint2('process_data', self)
        if ana.skipped:
            if ana.do_title:
                self.process_title(ana)
            return True

        if (self.isempty() and
            not self.get_figaxes().isempty() and
                not self._suppress):
            self.generate_artist()
#        app = self.get_app()
#        ax = self.get_container()
#        fig_ax = ax.figobj

        ishot = ana.ishot

        self._analysis_flag[ishot] = False
        col = color_order[ishot % len(color_order)]

        for x in range(ishot - self.num_child()+1):
            plot_options = self.getvar('plot_options')
            if self._plot_type == 'plot':
                kargs = plot_options['plot'][1].copy()
                kargs = set_auto_color_to_color(kargs, 'color', col)
                kargs = set_auto_color_to_color(kargs, 'markerfacecolor', col)
                kargs = set_auto_color_to_color(kargs, 'markeredgecolor', col)
                obj = FigPlot([0], [0],
                              *(plot_options['plot'][0]),
                              **kargs)
            elif self._plot_type == 'timetrace':
                kargs = plot_options['timetrace'][1].copy()
                kargs = set_auto_color_to_color(kargs, 'color', col)
                kargs = set_auto_color_to_color(kargs, 'markerfacecolor', col)
                kargs = set_auto_color_to_color(kargs, 'markeredgecolor', col)
                obj = TimeTrace([0], [0],
                                *(plot_options['timetrace'][0]),
                                **kargs)
            elif self._plot_type == 'stepplot':
                kargs = plot_options['stepplot'][1].copy()
                kargs = set_auto_color_to_color(kargs, 'color', col)
                kargs = set_auto_color_to_color(kargs, 'markerfacecolor', col)
                kargs = set_auto_color_to_color(kargs, 'markeredgecolor', col)
                obj = StepPlot([0], [0],
                               *(plot_options['stepplot'][0]),
                               **kargs)
            elif self._plot_type == 'contour':
                obj = FigContour([0, 1], [0, 1], np.arange(4).reshape(2, 2),
                                 *(plot_options['contour'][0]))
            elif self._plot_type == 'image':
                obj = FigImage(np.zeros([2, 2]))
            elif self._plot_type == 'surface':
                obj = FigSurface(np.zeros([2, 2]))
            elif self._plot_type == 'axspan':
                obj = FigAxspan([0, 1])
            elif self._plot_type == 'axline':
                obj = FigAxline(np.zeros([1]))
            elif self._plot_type == 'text':
                obj = FigText(0, 0, '')
            name = self._plot_type + str(self.num_child()+1)
            self.add_child(name, obj)
            obj.set_container_idx(self._container_idx)
            obj.set_suppress(True)
        if self.get_script_local():
            txt = self.make_script(ana.shot, short=True)
            debug_mode = viewer.debug_mode
            if txt is not None:
                try:
                    vars = viewer.get_globals_for_figmds(ana.shot)
#                  print vars
                    for key in vars:
                        ana.result[key] = vars[key]
                    for key in viewer.startup_values:
                        ana.result[key] = viewer.startup_values[key]
                    ana.result['shot'] = ana.shot
                    if self.getvar('posvars') is not None:
                        a = {}
                        b = {}
                        exec(self.getvar('posvars'), a, b)
                        for key in b:
                            ana.result[key] = b[key]
                    filepath = os.path.join(self.owndir(), script_file_name)
                    from ifigure.widgets.debugger_core import get_breakpoint
                    code = compile(txt, filepath, 'exec')
                    if (len(get_breakpoint(filepath)) != 0 and debug_mode):
                        app = wx.GetApp().TopWindow
                        se = app.script_editor
                        import threading
                        if not se.CheckDebuggerStatus():
                            se.QueueSEDRequest(code, viewer.g, ana.result,
                                               filepath)
#                          print threading.current_thread().name
#                          wx.Yield()
#                          time.sleep(3)
                        else:
                            se.RunSED(code, viewer.g, ana.result, filepath)
                    else:
                        exec(code, viewer.g, ana.result)
                except:
                    dprint1('error occured when processing data by script')
                    print('error occured when processing data by following script')
                    print('#####')
                    print(txt)
                    print('#####')
                    print(traceback.format_exc())
                    self.change_suppress(True, self.get_child(ishot))
                    return False

        if ana.do_title:
            self.process_title(ana)
        if self._plot_type == 'plot':
            self._update_plot(ana, ishot)  # , color_order)
        elif self._plot_type == 'timetrace':
            self._update_plot(ana, ishot)  # , color_order)
        elif self._plot_type == 'stepplot':
            self._update_plot(ana, ishot)  # , color_order)
        elif self._plot_type == 'contour':
            self._update_contour(ana, ishot)
        elif self._plot_type == 'image':
            self._update_image(ana, ishot)
        elif self._plot_type == 'surface':
            self._update_surface(ana, ishot)
        elif self._plot_type == 'axspan':
            self._update_axspan(ana, ishot)
        elif self._plot_type == 'axline':
            self._update_axline(ana, ishot)
        elif self._plot_type == 'text':
            self._update_text(ana, ishot)

        obj = self.get_child(idx=ishot)
        if obj.is_suppress():
            return False

        obj._data_extent = None
        try:
            if (self.get_figaxes() is not None):
                self.get_figaxes().adjust_axes_range()
        except:
            dprint1("failed in adjusting axes at postprocess_data, continuing...")
            pass
        return True

    def process_title(self, ana):
        ishot = ana.ishot
        obj = self.get_child(idx=ishot)
        if 'title' in ana.result:
            fig_ax = self.get_figaxes()
            if len(fig_ax._artists) == 0:
                v = fig_ax.getp('title_labelinfo')
                # print v
                if v[0] != ana.result['title']:
                    v[0] = ana.result['title']
                    fig_ax.set_title(v, None)
                    fig_ax.set_bmp_update(False)
#                v[0] = ana.result['title']
#                fig_ax.set_title(v, None)
            else:
                a = fig_ax._artists[0]
                v = fig_ax.get_title(a)
                if v[0] != ana.result['title']:
                    v[0] = ana.result['title']
                    fig_ax.set_title(v, a)
                    fig_ax.set_bmp_update(False)
        if obj is not None:
            self.change_suppress(False, obj)
        else:
            print('process tiltle error')
            print(ana)
            print(obj)
            print(ishot)

    def call_refresh_artist(self, ishot):
        obj = self.get_child(idx=ishot)

        # this happens if plot type is changed
        if obj is None:
            return

        if (not obj.is_suppress() and
                not self.get_figaxes().isempty()):
            obj.refresh_artist_data()
            if self._plot_type == 'plot':
                if len(obj._artists) != 0:
                    col = self._color_order[ishot % len(self._color_order)]
                    opt = self.getvar('plot_options')['plot'][1]
                    if opt['color'] == 'auto':
                        obj._artists[0].set_color(col)
                    if opt['markerfacecolor'] == 'auto':
                        obj._artists[0].set_markerfacecolor(col)
                    if opt['markeredgecolor'] == 'auto':
                        obj._artists[0].set_markeredgecolor(col)
            elif self._plot_type == 'timetrace':
                if len(obj._artists) != 0:
                    col = self._color_order[ishot % len(self._color_order)]
                    opt = self.getvar('plot_options')['timetrace'][1]
                    if opt['color'] == 'auto':
                        obj._artists[0].set_color(col)
                    if opt['markerfacecolor'] == 'auto':
                        obj._artists[0].set_markerfacecolor(col)
                    if opt['markeredgecolor'] == 'auto':
                        obj._artists[0].set_markeredgecolor(col)

    def _en_numpy(self, v, name, s=True):
        from ifigure.utils.cbook import isiterable, isndarray
        if not name in v:
            return None, s and False
        if isiterable(v[name]) and not isndarray(v[name]):
            try:
                v[name] = np.array(v[name])
            except:
                return v[name], True and s
        return v[name], True and s

    def _update_plot(self, ana, ishot):  # , color_order):
        x, s = self._en_numpy(ana.result, 'x')
        y, s = self._en_numpy(ana.result, 'y', s)
        if not s:
            dprint2('Incorrect data to update plot :'+self.get_full_path())
            dprint2(ana.result.__repr__())
            obj = self.get_child(idx=ishot)
            self.change_suppress(True, obj)
            return
        try:
            obj = self.get_child(idx=ishot)
            if y is not None:
                if x is not None:
                    if len(x.shape) > 1:
                        x = x.flatten()
                    if len(y.shape) > 1:
                        y = y.flatten()
                    if x.size < y.size:
                        y = y[0:x.size]
                    if x.size > y.size:
                        x = x[0:y.size]
                    if x.ndim == 0:
                        x = np.array([x])
                    if y.ndim == 0:
                        y = np.array([y])
                    obj.setvar('x', x)
                    obj.setvar('y', y)
                    if 'x_catalog' in ana.result:
                        obj.setvar('x_catalog', ana.result['x_catalog'])
                    else:
                        obj.delvar('x_catalog')
                    if 'y_catalog' in ana.result:
                        obj.setvar('y_catalog', ana.result['y_catalog'])
                    else:
                        obj.delvar('y_catalog')

#               obj._is_decimate = True
                    obj.setp('use_var', True)
                else:
                    if len(y.shape) > 1:
                        y = y.flatten()
                    x = obj.getvar('x')
                    if x.size < y.size:
                        y = y[0:x.size]
                    if x.size > y.size:
                        x = x[0:y.size]
                    if y.ndim == 0:
                        y = np.array([y])
                    obj.setvar('y', y)
                    obj.setp('use_var', True)
                self.change_suppress(False, obj)
#           obj.set_suppress(False)
                obj._mpl_cmd = 'plot'
                from ifigure.mto.fig_plot import FigPlot, TimeTrace, StepPlot
                if (not isinstance(obj, TimeTrace) and
                        not isinstance(obj, StepPlot)):
                    if 'yerr' in ana.result:
                        yerr, s = self._en_numpy(ana.result, 'yerr')
                        obj._mpl_cmd = 'errorbar'
                        obj.setvar('yerr', yerr)
                    if 'xerr' in ana.result:
                        xerr, s = self._en_numpy(ana.result, 'xerr')
                        obj._mpl_cmd = 'errorbar'
                        obj.setvar('xerr', xerr)
                if self._suppress:
                    self.set_suppress(True)
                self._analysis_flag[ishot] = True
            else:
                logging.basicConfig(level=logging.DEBUG)
                logging.exception('session failed in visualization step')
                self.change_suppress(True, obj)
                ana.print_jobs()
                return
        except Exception:
            logging.basicConfig(level=logging.DEBUG)
            logging.exception('error in post processing')
            print(ana.result)
            obj.setvar('x', [0])
            obj.setvar('y', [0])
            obj.setp('use_var', True)
            self.change_suppress(True, obj)
#           if not obj.is_suppress():
#               obj.refresh_artist_data()

    def _update_axspan(self, ana, ishot):
        obj = self.get_child(idx=ishot)
        obj.setp('use_var', True)
        obj.setvar('x', [])
        obj.setvar('y', [])
        sup = True
        if 'x' in ana.result:  # x direction
            x, s = self._en_numpy(ana.result, 'x')
            obj.setvar('x', x)
            obj._data_extent = None
            sup = False
        if 'y' in ana.result:  # y direction
            y, s = self._en_numpy(ana.result, 'y')
            obj.setvar('y', y)
            obj._data_extent = None
            sup = False
        obj.set_suppress(sup)
        if self._suppress:
            self.set_suppress(True)
        self._analysis_flag[ishot] = not sup

    def _update_axline(self, ana, ishot):
        # print ana
        obj = self.get_child(idx=ishot)
        obj.setp('use_var', True)
        obj.setvar('x', [])
        obj.setvar('y', [])
        sup = True
        if 'x' in ana.result:  # x direction
            x, s = self._en_numpy(ana.result, 'x')
            obj.setvar('x', x)
            obj._data_extent = None
            sup = False
        if 'y' in ana.result:  # y direction
            y, s = self._en_numpy(ana.result, 'y')
            # print 'y', s
            obj.setvar('y', y)
            obj._data_extent = None
            sup = False
        obj.set_suppress(sup)
        if self._suppress:
            self.set_suppress(True)
        self._analysis_flag[ishot] = not sup

    def _update_surface(self, ana, ishot):
        self._update_contour(ana, ishot)

    def _update_image(self, ana, ishot):
        self._update_contour(ana, ishot)

    def _update_contour(self, ana, ishot):
        x, s = self._en_numpy(ana.result, 'x')
        y, s = self._en_numpy(ana.result, 'y', s)
        z, s = self._en_numpy(ana.result, 'z', s)
        if not s:
            dprint2('Incorrect data to update contour/image :' +
                    self.get_full_path())
            dprint2(ana.result.__repr__())
            self.suppress_ishot(ishot)
#           obj = self.get_child(idx=ishot)
#           self.change_suppress(True, obj)
            return
        try:
            obj = self.get_child(idx=ishot)
            if z is not None:
                if y is None:
                    y = np.arange(z.shape[0])
                if x is None:
                    x = np.arange(z.shape[1])
                obj.setvar('x', x)
                obj.setvar('y', y)
                obj.setvar('z', z)
                obj.setp('use_var', True)
                obj.set_suppress(False)
                if self._suppress:
                    self.set_suppress(True)
                self._analysis_flag[ishot] = True
            else:
                logging.basicConfig(level=logging.DEBUG)
                logging.exception('session failed in visualization step')
                self.change_suppress(True, obj)
                ana.print_jobs()
                return
        except:
            logging.basicConfig(level=logging.DEBUG)
            logging.exception('error in post processing')
            obj.setvar('z', np.zeros([2, 2]))
            obj.setvar('x', [0, 1])
            obj.setvar('y', [0, 1])
            obj.setp('use_var', True)
            self.change_suppress(True, obj)
#           if not obj.is_suppress():
#              obj.refresh_artist_data()

    def _update_text(self, ana, ishot):
        try:
            obj = self.get_child(idx=ishot)
            obj.setvar('s', ana.result['s'])
            x, y = obj.get_gp_point(0)
            if 'x' in ana.result:
                obj.setvar('x', ana.result['x'])
            else:
                obj.setvar('x', x)
            if 'y' in ana.result:
                obj.setvar('y', ana.result['y'])
            else:
                obj.setvar('y', y)
            obj.setp('use_var', True)
            if self._suppress:
                self.set_suppress(False)
            self._analysis_flag[ishot] = True
        except:
            logging.basicConfig(level=logging.DEBUG)
            logging.exception('error in post processing')
            obj.setvar('s', '')
#            obj.setvar('x', )
#            obj.setvar('y', [0,1])
            obj.setp('use_var', True)
            self.change_suppress(True, obj)

    def active_plots(self, l):
        if len(self._d) > l:
            for plot in self._d[l:]:
                self.change_suppress(True, plot)

    def eval_mdstitle(self):
        title = self.getvar('title')
        return title

    def set_mdssource(self, value, a):
        # a is not used
        for name, val in value:
            if name in self.getvar("mdsvars"):
                self.getvar("mdsvars")[name] = str(val)
            elif self.hasvar(name):
                self.setvar(name, str(val))

    def get_mdssource(self, a):
        # a is not used
        #        return {"data":
        #               [self.getvar("experiment"),
        #                self.getvar("default_node"),
        #                self.getvar("mdsvars")["x"],
        #                self.getvar("mdsvars")["y"],
        #                self.getvar("mdsvars")["z"],
        #                self.getvar("mdsvars")["xerr"],
        #                self.getvar("mdsvars")["yerr"],
        #                self.getvar("title"),
        #                self.getvar("update")],
        return {"figmds": self,
                "posvars": self.getvar("posvars")}


#    def handle_axes_change(self,evt = None):
#        for name, child in self.get_children():
#            child.handle_axes_change(evt)

    def apply_mdsrange(self, x=None, y=None):
        if x is None:
            x = [None, None]
        if not self._default_xyrange['flag'][0]:
            try:
                x = [float(m) for m in self._default_xyrange['tdi'][0]]
            except:
                pass
        if y is None:
            y = [None, None]
        if not self._default_xyrange['flag'][1]:
            try:
                y = [float(m) for m in self._default_xyrange['tdi'][1]]
            except:
                pass
        self._default_xyrange['value'] = (x, y)

    def set_mdsrange(self, value, a):
        self._default_xyrange['flag'] = (value[0][0], value[1][0])
        self._default_xyrange['tdi'] = (value[0][1], value[1][1])
        self.apply_mdsrange()
        return

    def get_mdsrange(self, a):
        v = ((self._default_xyrange['flag'][0],
              self._default_xyrange['tdi'][0]),
             (self._default_xyrange['flag'][1],
              self._default_xyrange['tdi'][1]))

        return v

    def set_mdsfiguretype(self, value, a=None):
        if self._plot_type != value:
            self.erase_mdsdata()
        ovalue = self._plot_type
        self._plot_type = value
        if self._session_editor is not None:
            self._session_editor.set_required_variables(
                required_variables[self._plot_type])

        figaxes = self.get_figaxes()

        threed_item = ['surface']
        flag1 = False
        if (ovalue in threed_item and
                not value in threed_item):
            flag1 = True
        if (not ovalue in threed_item and
                value in threed_item):
            flag1 = True

#        print value, ovalue, flag1
        if not flag1:
            return

        flag = False
        for obj in figaxes.walk_tree():
            if isinstance(obj, FigMds):
                if obj._plot_type in threed_item:
                    flag = True
                    break
        figaxes.set_3d(flag)
        ifigure.events.SendPVDrawRequest(self)

    def get_mdsfiguretype(self, a=None):
        return self._plot_type

    def get_script_local(self, a=None):
        v = self.get_figbook().getvar('mdsscript_main')
        if v is not None:
            return v
        return True

    def change_suppress(self, value, obj):
        if value == obj._suppress:
            return
#            if not value and not obj.isempty(): return
#            if value and obj.isempty(): return
        obj.set_suppress(value)

    def suppress_ishot(self, ishot):
        obj = self.get_child(ishot)
#        print 'suppressing ', obj, ishot
        if obj is not None:
            obj.set_suppress(True)
            obj.realize()

    def set_suppress_all(self, value):
        for name, child in self.get_children():
            child.set_suppress(value)

    def is_all_suppressed(self):
        return all([child._suppress for name, child in self.get_children()])

    def eval_shotstr(self, txt, prevshot, current):
        mode = ''
        if txt.startswith('*'):
            txt = txt[1:]
            mode = '*'
        if txt.startswith('-'):
            txt = 'p'+txt
        elif txt.startswith('+'):
            txt = 'p'+txt

        lc = {'c': current, 'p': prevshot, 'm': -1}
        if self.getvar("posvars") is not None:
            exec(self.getvar("posvars"), {}, lc)
        if txt.strip() == '':
            return 0, prevshot
        try:
            val = eval(txt, {}, lc)
        except:
            print('Failed to evaluate shot text')
            print(('shot text', txt))
            print(('name space', lc))
            # print traceback.format_exc()
            return None, None
        if 0 < val < 1000:
            val = (prevshot/1000)*1000 + val
        else:
            prevshot = val
        if val == -1:
            val = 'm'
            if mode == '*':
                val = 'n'
            prevshot = current
        else:
            if mode == '*':
                val = -abs(val)
#        print val, prevshot
        return val, prevshot

    def save2(self, fid=None, olist=None):
        return super(FigMds, self).save2(fid=fid, olist=olist)
        if olist is None:
            olist = []

        h2 = {"name": self.__class__.__name__,
              "module": self.__class__.__module__,
              "num_child": self.num_child(),
              "sname": self.name,
              "id": self.id,
              "var0": self._var0,
              "var": self._var,
              "note": self._note,
              "format": 2}
#        if self._save_mode == 1:
#            h2["num_child"] = 0
#        print h2
        if fid is not None:
            pickle.dump(h2, fid)
        c = olist.count(self.id)
        if c == 0:
            data = self.save_data2({})
            pickle.dump(data, fid)
            olist.append(self.id)
        else:
            pass
#        if self._save_mode == 1:
#            return olist
        for name, child in self.get_children():
            olist = child.save2(fid, olist)
        return olist

    def save_data2(self, data):
        # the first element is version code
        param = {"_plot_type":          self._plot_type,
                 "_default_xyrange":    self._default_xyrange,
                 "_var_mask":    self._var_mask}
        if self.get_figbook().has_child('dwglobal'):
            dwglobal = {'globalsets': self.get_figbook(
            ).dwglobal.getvar('globalsets')}
        else:
            dwglobal = None
        data['FigMds'] = (1, {}, param, dwglobal)
        data = super(FigMds, self).save_data2(data)
        return data

    def load_data2(self, data):
        def check_default_options(mode, name, v):
            if not mode in opt:
                opt[mode] = def_plot_options[mode]
            opt[mode][1][name] = v
        if 'FigMds' in data:
            val = data['FigMds'][1]
            for key in val:
                self.setp(key, val[key])
            if len(data['FigMds']) > 2:
                param = data['FigMds'][2]
                for key in param:
                    if hasattr(self, key):
                        setattr(self, key, param[key])
            if len(data['FigMds']) > 3:
                self.setp('loaded_dwglobal', data['FigMds'][3])

        super(FigMds, self).load_data2(data)

        # adjustment for backword compatibility
        opt = self.getvar('plot_options')
        check_default_options('plot', 'color', 'auto')
        check_default_options('plot', 'markerfacecolor', 'auto')
        check_default_options('plot', 'markeredgecolor', 'auto')
        check_default_options('timetrace', 'color', 'auto')
        check_default_options('timetrace', 'markerfacecolor', 'auto')
        check_default_options('timetrace', 'markeredgecolor', 'auto')
        check_default_options('stepplot', 'color', 'auto')
        check_default_options('stepplot', 'markerfacecolor', 'auto')
        check_default_options('stepplot', 'markeredgecolor', 'auto')

    def write_script(self, script):
        if not self.has_owndir():
            self.mk_owndir()
        filename = 'mdsscript.py'
        self.setvar('pathmode', 'owndir')
        self.setvar('path', filename)
        fname = os.path.join(self.owndir(),
                             self.getvar('path'))
        from ifigure.mdsplus.fig_mds import write_scriptfile
        write_scriptfile(fname, script)

    def remove_script(self):
        if (self.has_owndir() and
                self.hasvar('path')):
            fname = os.path.join(self.owndir(), self.getvar('path'))
            if os.path.exists(fname):
                os.remove(fname)
                self.delvar('pathmode')
                self.delvar('path')

    def picker_a(self, a, evt):
        s = [child._suppress for name, child in self.get_children()]
        if len(s) == 0 or all(s):
            if self._artists[0].contains(evt):
                return False, {}
            else:
                return True, {"child_artist": self._artists[0]}
        return super(FigMds, self).picker_a(a, evt)



import ifigure.utils.pickle_wrapper as pickle
import select
import time


class PickledPipe(object):
    '''
       PickledPipe(a, b)
       a: buffer for reading
       b: buffer for writing
    '''

    def __init__(self, in_f, out_f):
        self.in_f = in_f
        self.out_f = out_f

    def send(self, data):
        pickle.dump(data, self.out_f)
        self.out_f.flush()

    def recv(self, nowait=True):
        #        print self._check_readbuf()
        if self._check_readbuf() or not nowait:
            try:
                return pickle.load(self.in_f)
            except:
                return {'error message': ['pickle communicaiton error']}
        else:
            return None

    def _check_readbuf(self):
        #        print select.select([self.in_f],[],[], 0)[0]
        return self.in_f in select.select([self.in_f], [], [], 0)[0]
